
from epyk.core.css import Properties


class Data(Properties.CssMixin):
  def __init__(self, css_ovrs, selector):
    self.htmlObj = self
    self._content = css_ovrs or {}
    self.__selector = selector
    self.has_changed = False

  def css(self, k, v=None, important=False, change=True):
    """

    :param k:
    :param v:
    :param important:
    :param change:

    :return:
    """
    if change:
      self.has_changed = True
    if not isinstance(k, dict) and v is not None:
      k = {k: v}
    for attr, val in k.items():
      if important:
        val = "%s !IMPORTANT" % val
      self._content[attr] = val
    return self

  def clear(self):
    """
    Clear all the CSS attributes defined for this class

    :return: Self to allow the chaining
    """
    self.has_changed = True
    self._content = {}
    return self

  @property
  def selector(self):
    return self.__selector

  def __str__(self):
    if self._content:
      return "{%s;}" % ";".join(["%s: %s " % (k, v) for k, v in self._content.items()])

    return ""


class Selector(object):
  def __init__(self, selector_ovrs):
    map_fncs = {"parent": "parent_element", "child": "sub_element"}
    self.__this, self._suffix = selector_ovrs["classname"], ''
    del selector_ovrs["classname"]
    for k, v in selector_ovrs.items():
      getattr(self, map_fncs.get(k, k))(v)
    self.__element = False

  def elements(self, element_types):
    """
    Selects all <div> elements and all <p> elements

    Related Pages:

			https://www.w3schools.com/cssref/sel_element_comma.asp

    :param element_types:
    :return:
    """
    if not isinstance(element_types, list):
      element_types = [element_types]
    if self.__element:
      self.__this = "%s,%s" % (",".join(element_types), self.__this)
    else:
      self.__this = "%s%s" % (",".join(element_types), self.__this)
      self.__element = True
    return self

  def sub_element(self, element, direct_parent=False, class_name=None):
    """
    Selects all <p> elements inside <div> elements

    Related Pages:

			https://www.w3schools.com/cssref/sel_element_element.asp

    :param element:
    :return:
    """
    if direct_parent:
      self.__this = "%s > %s" % (self.__this, element)
    else:
      self.__this = "%s %s" % (self.__this, element)
    return self

  def parent_element(self, element, direct_parent=False, class_name=None):
    """
    Selects all <p> elements inside <div> elements

    Related Pages:

			https://www.w3schools.com/cssref/sel_element_element.asp

    :param element:
    :return:
    """
    if direct_parent:
      self.__this = "%s > %s" % (element, self.__this)
    else:
      self.__this = "%s %s" % (element, self.__this)
    return self

  def with_next_element(self, element, class_name=None):
    """
    Selects all <p> elements that are placed immediately after <div> elements

    Related Pages:

			https://www.w3schools.com/cssref/sel_element_pluss.asp

    :param element:
    :return:
    """
    self.__this = "%s + %s" % (self.__this, element)
    return self

  def with_prev_element(self, element, class_name=None):
    """
    Selects every <ul> element that are preceded by a <p> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_gen_sibling.asp

    :param element:
    :return:
    """
    self.__this = "%s ~ %s" % (self.__this, element)
    return self

  def join_class(self, class_name):
    """
    Selects all elements with both name1 and name2 set within its class attribute

    Documentation

    :param class_name:
    :return:
    """
    self.__this = "%s.%s" % (self.__this, class_name)
    return self

  def add_class(self, class_name):
    """
    Selects all elements with name2 that is a descendant of an element with name1

    :param class_name:
    :return:
    """
    self.__this = "%s .%s" % (self.__this, class_name)
    return self

  def sub_class(self, class_name, direct_parent=False):
    """
    Selects all <p> elements inside <div> elements

    Related Pages:

			https://www.w3schools.com/cssref/sel_element_element.asp

    :param class_name:
    :return:
    """
    if direct_parent:
      self.__this = "%s > .%s" % (self.__this, class_name)
    else:
      self.__this = "%s .%s" % (self.__this, class_name)
    return self

  def element(self, bool=None):
    """

    :param element:
    :return:
    """
    if bool is None:
      return self.__element

    self.__element = bool
    return self

  def add_element_id(self, htmlObj):
    """
    Selects the element with id="firstname"

    Related Pages:

			https://www.w3schools.com/cssref/sel_id.asp

    :param htmlObj:
    :return:
    """
    self.__this = "%s%s" % (htmlObj.htmlCode, self.__this)
    return self

  def not_element(self, element):
    """
    Selects every element that is not a <p> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_not.asp

    :param element:
    :return:
    """
    self.__this = "%s:not(%s)" % (self.__this, element)
    return self

  def first_child(self):
    """
    Selects every <p> element that is the first child of its parent

    Related Pages:

			https://www.w3schools.com/cssref/sel_firstchild.asp

    :return:
    """
    self.__this = "%s:first-child" % self.__this
    return self

  def first_letter(self):
    """
    Selects the first letter of every <p> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_firstletter.asp

    :return:
    """
    self.__this = "%s::first-letter" % self.__this
    return self

  def first_line(self):
    """
    Selects the first line of every <p> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_firstline.asp

    :return:
    """
    self.__this = "%s::first-line" % self.__this
    return self

  def first_of_type(self):
    """
    Selects every <p> element that is the first <p> element of its parent

    Related Pages:

			https://www.w3schools.com/cssref/sel_first-of-type.asp

    :return:
    """
    self.__this = "%s:first-of-type" % self.__this
    return self

  def in_range(self):
    """
    Selects input elements with a value within a specified range

    Related Pages:

			https://www.w3schools.com/cssref/sel_in-range.asp

    :return:
    """
    self.__this = "%s:in-range" % self.__this
    return self

  def last_child(self):
    """
    Selects every <p> element that is the last child of its parent

    Related Pages:

			https://www.w3schools.com/cssref/sel_last-child.asp

    :return:
    """
    self.__this = "%s:last-child" % self.__this
    return self

  def last_of_type(self):
    """
    Selects every <p> element that is the last <p> element of its parent

    Related Pages:

			https://www.w3schools.com/cssref/sel_last-of-type.asp

    :return:
    """
    self.__this = "%s:last-of-type" % self.__this
    return self

  def link(self):
    """
    Selects all unvisited links

    Related Pages:

			https://www.w3schools.com/cssref/sel_link.asp

    :return:
    """
    self.__this = "%s:link" % self.__this
    return self

  def nth_child(self, n):
    """
    Selects every <p> element that is the second child of its parent

    Related Pages:

			https://www.w3schools.com/cssref/sel_nth-child.asp

    :param n:
    :return:
    """
    self.__this = "%s:nth-child(%s)" % (self.__this, n)
    return self

  def nth_last_child(self, n):
    """
    Selects every <p> element that is the second child of its parent, counting from the last child

    Related Pages:

			https://www.w3schools.com/cssref/sel_nth-last-child.asp

    :param n:
    :return:
    """
    self.__this = "%s:nth-last-child(%s)" % (self.__this, n)
    return self

  def nth_last_of_type(self, n):
    """
    Selects every <p> element that is the second <p> element of its parent, counting from the last child

    Related Pages:

			https://www.w3schools.com/cssref/sel_nth-last-of-type.asp

    :param n:
    :return:
    """
    self.__this = "%s:nth-last-of-type(%s)" % (self.__this, n)
    return self

  def nth_of_type(self, n):
    """
    Selects every <p> element that is the second <p> element of its parent

    Related Pages:

			https://www.w3schools.com/cssref/sel_nth-of-type.asp

    :param n:
    :return:
    """
    self.__this = "%s:nth-of-type(%s)" % (self.__this, n)
    return self

  def only_of_type(self):
    """
    Selects every <p> element that is the only <p> element of its parent

    Related Pages:

			https://www.w3schools.com/cssref/sel_only-of-type.asp

    :return:
    """
    self.__this = "%s:only-of-type" % self.__this
    return self

  def only_child(self):
    """
    Selects every <p> element that is the only child of its parent

    Related Pages:

			https://www.w3schools.com/cssref/sel_only-child.asp

    :return:
    """
    self.__this = "%s:only-child" % self.__this
    return self

  def placeholder(self):
    """
    Selects input elements with placeholder text

    Related Pages:

			https://www.w3schools.com/cssref/sel_placeholder.asp

    :return:
    """
    self.__this = "%s::placeholder" % self.__this
    return self

  def read_only(self):
    """
    Selects input elements with the "readonly" attribute specified

    Related Pages:

			https://www.w3schools.com/cssref/sel_read-only.asp

    :return:
    """
    self.__this = "%s:read-only" % self.__this
    return self

  def read_write(self):
    """
    Selects input elements with the "readonly" attribute NOT specified

    Related Pages:

			https://www.w3schools.com/cssref/sel_read-write.asp

    :return:
    """
    self.__this = "%s:read-write" % self.__this
    return self

  def suffix(self, data):
    self._suffix = "%s " % data.strip()

  def __str__(self):
    return self.__this


class Style(object):
  classname, classnames = None, None

  def __init__(self, rptObj, css_ovrs=None, selector_ovrs=None, html_id=None):
    self.rptObj, self.html_id, self.cls_ref = rptObj, html_id, None
    css_ovrs = css_ovrs or {}
    self.__keyframes, self.__media = {}, {}
    selector_ids = dict(getattr(self, '_selectors', {}))
    if self.classname is None:
      self.classname = self.__class__.__name__.lower()
    if self.classnames is not None:
      self.classname = " .".join(self.classnames)
    if self.classname == False:
      selector_ids["classname"], self.classname = ("%s", "")
    elif self.classname.startswith('::'):
      selector_ids["classname"], self.classname = ("%s", self.classname)
    else:
      selector_ids["classname"], self.classname = (".%s", self.classname)
    if getattr(self, '_selector', None) is not None:
      self.classname = self._selector
    if selector_ovrs is not None:
      self.__has_changed = True
      selector_ids.update(selector_ovrs)
    self.__selector = Selector(selector_ids)

    self.__attrs, self.__hover, self.__focus = dict(getattr(self, '_attrs', {})), dict(getattr(self, '_hover', {})), dict(getattr(self, '_focus', {}))
    self.__checked, self.__disabled, self.__empty = dict(getattr(self, '_checked', {})), dict(getattr(self, '_disabled', {})), dict(getattr(self, '_empty', {}))
    self.__enabled, self.__invalid, self.__valid = dict(getattr(self, '_enabled', {})), dict(getattr(self, '_invalid', {})), dict(getattr(self, '_valid', {}))
    self.__visited, self.__after, self.__before = dict(getattr(self, '_visited', {})), dict(getattr(self, '_after', {})), dict(getattr(self, '_before', {}))
    self.__active = dict(getattr(self, '_active', {}))

    # More bespoke items
    self.__webkit_slider_thumb = dict(getattr(self, '_webkit_slider_thumb', {}))
    self.__internal_props = ["attrs", "hover", "focus", "checked", "valid", "disabled", "empty", "enabled", "invalid", "active"]
    self.customize()
    for k in self.__internal_props:
      s = getattr(self, k)
      s.has_changed = False
    self.__has_changed = False
    if css_ovrs:
      for k, v in css_ovrs.items():
        if k in self.__internal_props:
          getattr(self, k).css(k, v)
        else:
          self.attrs.css(k, v)

  def customize(self):
    pass

  def transition(self, attribute, duration=2, delay=None, iteration=None, timing_fnc=None):
    """

    :param attribute:
    :param duration:
    :param delay:
    :param iteration:
    :param timing_fnc:
    """
    css_transition = {"transition-property": attribute, "transition-duration": "%ss" % duration}
    if delay:
      css_transition["transition-delay"] = "%ss" % delay
    if iteration:
      css_transition["transition-iteration-count"] = iteration
    if timing_fnc is not None:
      if timing_fnc not in ["ease", "linear", "ease-in", "ease-out", "ease-in-out"] and not timing_fnc.startswith("cubic-bezier"):
        raise Exception("%s missing from the list" % timing_fnc)

      css_transition["transition-timing-function"] = timing_fnc
    # Add the -webkit- prefix for compatibility with some browsers
    safari_css = dict([("-webkit-%s" % k, v) for k, v in css_transition.items()])
    css_transition.update(safari_css)
    self.css(css_transition, change=False)
    return self

  def animation(self, name=None, attrs=None, duration=2, delay=None, iteration='infinite', timing_fnc=None, effect=None, fill_mode=None):
    """
    The @keyframes rule specifies the animation code.

    The animation is created by gradually changing from one set of CSS classes to another.

    Example
    rptObj.ui.button("Ok").style.css_class.animation('test', {
      "from": {"border-color": "white"},
      "to": {"border-color": "red"},
    })

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_animation-keyframes.asp
    https://www.w3schools.com/css/css3_animations.asp

    :param effect: Effect Class.
    :param name: String. Required. Defines the name of the animation.
    :param attrs: String. Required. Percentage of the animation duration.
    :param duration:
    :param delay:
    :param iteration:
    :param timing_fnc:
    :param fill_mode: String Specify the fill mode (whether the style should go back to its original position or something else etc...
    """
    name = self.keyframes(name, attrs, effect)
    css_animation = {"animation-name": name, "animation-duration": "%ss" % duration}
    if delay:
      css_animation["animation-delay"] = "%ss" % delay
    if iteration:
      css_animation["animation-iteration-count"] = iteration
    if fill_mode:
      css_animation['animation-fill-mode'] = fill_mode
    if timing_fnc is not None:
      if timing_fnc not in ["ease", "linear", "ease-in", "ease-out", "ease-in-out"] and not timing_fnc.startswith(
        "cubic-bezier"):
        raise Exception("%s missing from the list" % timing_fnc)

      css_animation["animation-timing-function"] = timing_fnc
    # Add the -webkit- prefix for compatibility with some browsers
    safari_css = dict([("-webkit-%s" % k, v) for k, v in css_animation.items()])
    css_animation.update(safari_css)
    self.css(css_animation, change=False)
    return self

  @property
  def selector(self):
    return self.__selector

  @property
  def attrs(self):
    if self.__attrs is None or isinstance(self.__attrs, dict):
      self.__attrs = Data(self.__attrs or {}, self.__selector)
    return self.__attrs

  @property
  def hover(self):
    """
    Selects links on mouse over

    Related Pages:

			https://www.w3schools.com/cssref/sel_hover.asp

    :return:
    """
    if self.__hover is None or isinstance(self.__hover, dict):
      self.__hover = Data(self.__hover or {}, self.selector)
    return self.__hover

  @property
  def checked(self):
    """
    Selects every checked <input> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_checked.asp

    :return:
    """
    if self.__checked is None or isinstance(self.__checked, dict):
      self.__checked = Data(self.__checked or {}, self.selector)
    return self.__checked

  @property
  def disabled(self):
    """
    Selects every disabled <input> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_disabled.asp

    :return:
    """
    if self.__disabled is None or isinstance(self.__disabled, dict):
      self.__disabled = Data(self.__disabled or {}, self.selector)
    return self.__disabled

  @property
  def focus(self):
    """
    Selects the input element which has focus

    Related Pages:

			https://www.w3schools.com/cssref/sel_focus.asp

    :return:
    """
    if self.__focus is None or isinstance(self.__focus, dict):
      self.__focus = Data(self.__focus or {}, self.selector)
    return self.__focus

  @property
  def empty(self):
    """
    Selects every <p> element that has no children (including text nodes)

    Related Pages:

			https://www.w3schools.com/cssref/sel_empty.asp

    :return:
    """
    if self.__empty is None or isinstance(self.__empty, dict):
      self.__empty = Data(self.__empty or {}, self.selector)
    return self.__empty

  @property
  def enabled(self):
    """
    Selects every enabled <input> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_enabled.asp

    :return:
    """
    if self.__enabled is None or isinstance(self.__enabled, dict):
      self.__enabled = Data(self.__enabled or {}, self.selector)
    return self.__enabled

  @property
  def invalid(self):
    """
    Selects all input elements with an invalid value

    Related Pages:

			https://www.w3schools.com/cssref/sel_invalid.asp

    :return:
    """
    if self.__invalid is None or isinstance(self.__invalid, dict):
      self.__invalid = Data(self.__invalid or {}, self.selector)
    return self.__invalid

  @property
  def valid(self):
    """
    Selects all input elements with a valid value

    Related Pages:

			https://www.w3schools.com/cssref/sel_valid.asp

    :return:
    """
    if self.__valid is None or isinstance(self.__valid, dict):
      self.__valid = Data(self.__valid or {}, self.selector)
    return self.__valid

  @property
  def visited(self):
    """
    Selects all visited links

    Related Pages:

			https://www.w3schools.com/cssref/sel_visited.asp

    :return:
    """
    if self.__visited is None or isinstance(self.__visited, dict):
      self.__visited = Data(self.__visited or {}, self.selector)
    return self.__visited

  @property
  def before(self):
    """
    Insert something before the content of each <p> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_before.asp

    :return:
    """
    if self.__before is None or isinstance(self.__before, dict):
      self.__before = Data(self.__before or {}, self.selector)
    return self.__before

  @property
  def after(self):
    """
    Insert something after the content of each <p> element

    Related Pages:

			https://www.w3schools.com/cssref/sel_after.asp

    :return:
    """
    if self.__after is None or isinstance(self.__after, dict):
      self.__after = Data(self.__after or {}, self.selector)
    return self.__after

  @property
  def webkit_slider_thumb(self):
    """
    :return:
    """
    if self.__webkit_slider_thumb is None or isinstance(self.__webkit_slider_thumb, dict):
      self.__webkit_slider_thumb = Data(self.__webkit_slider_thumb or {}, self.selector)
    return self.__webkit_slider_thumb

  @property
  def active(self):
    """
    Selects the active link

    Related Pages:

			https://www.w3schools.com/cssref/sel_active.asp

    :return:
    """
    if self.__active is None or isinstance(self.__active, dict):
      self.__active = Data(self.__active or {}, self.selector)
    return self.__active

  @property
  def has_changed(self):
    """

    :return:
    """
    for e in self.__internal_props:
      s = getattr(self, e)
      self.__has_changed |= s.has_changed
    return self.__has_changed

  def css(self, key, value=None, important=False, change=True):
    """

    :param key:
    :param value:
    :param important:
    :param change:
    :return:
    """
    return self.attrs.css(key, value, important, change)

  def media(self, attrs, rule=None, mediatype=None, mediafeature=None, change=True):
    """
    The @media is used in media queries to apply different styles for different media types/devices.

    Example
    rptObj.style.media({"body": {"background-color": "lightblue"}}, only, screen,
    {'and': [{'height': '100px'}, {'min-width': '600px'}]})

    The first key of the attributes can be an Epyk html object

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_mediaquery.asp
    :param attrs: String. Required. Percentage of the animation duration.
    :param rule: not or only or and see documentation for more info
    :param media_type: the media to which the rule will need to be applied
    :param mediafeature: Media features provide more specific details to media queries
    """

    if change:
      self.__has_changed = True
    media_props = []
    if rule is not None:
      if rule in ['only', 'not'] and not mediatype:
        raise Exception('You need to specify a mediatype when using rules not or only')

      media_props.extend([rule, mediatype])
      if mediafeature:
        for op, m_features in mediafeature.items():
          for feature in m_features:
            features = ['(%s: %s)' % (k, v) for k, v in feature.items()]
          media_props.extend([op, ('%s ' % op).join(features)])
    name = ' '.join(media_props)
    self.__media[name] = attrs
    return name

  def keyframes(self, name, attrs, effects=None, change=True):
    """
    The @keyframes rule specifies the animation code.

    The animation is created by gradually changing from one set of CSS styles to another.

    Example
    rptObj.style.keyframes("test", {
      "50%": {"transform": "scale(1.5, 1.5)", "opacity": 0},
      "99%": {"transform": "scale(0.001, 0.001)", "opacity": 0},
      "100%": {"transform": "scale(0.001, 0.001)", "opacity": 1},
    })

    Related Pages:

			https://www.w3schools.com/cssref/css3_pr_animation-keyframes.asp

    :param effects: Effect Class.
    :param name: String. Required. Defines the name of the animation.
    :param attrs: String. Required. Percentage of the animation duration.
    """
    if change:
      self.__has_changed = True
    if effects is not None:
      name = effects.__class__.__name__
      attrs = effects.get_attrs()
    self.__keyframes[name] = attrs
    return name

  def get_ref(self):
    """

    :return:
    """
    if self.has_changed and self.cls_ref in [None, self.classname]:
      # dedicated unique ID if it is not the original style
      if self.html_id is not None:
        self.cls_ref = "%s_%s" % (self.classname, self.html_id)
      else:
        self.cls_ref = self.classname
    elif not self.has_changed:
      self.cls_ref = self.classname
    return self.cls_ref

  def __str__(self):
    style = []
    cls_reference = self.get_ref()
    for e in self.__internal_props:
      s = getattr(self, e)
      css_id = str(s.selector) % cls_reference
      if str(s):
        if e == "attrs":
          style.append("%s %s%s" % (css_id.strip(), s.selector._suffix, s))
        else:
          style.append("%s:%s %s%s" % (css_id.strip(), e, s.selector._suffix, s))
    map_cls_ref = {'-webkit-slider-thumb': 'webkit_slider_thumb'}
    for e in ['after', 'before', '-webkit-slider-thumb']:
      s = getattr(self, map_cls_ref.get(e, e))
      css_id = str(s.selector) % cls_reference
      if str(s):
        style.append("%s::%s %s" % (css_id.strip(), e, s))
    if self.__keyframes:
      for name, k_attrs in self.__keyframes.items():
        style.append("@keyframes %s {" % name)
        for k, v_dict in k_attrs.items():
          style.append("  %s {%s; }" % (k, "; ".join(["%s: %s" % (i, ", ".join(j)) if isinstance(j, list) else "%s: %s" % (i, j) for i, j in v_dict.items()])))
        style.append("}")
    if self.__media:
      for name, m_attrs in self.__media.items():
        style.append("@media %s {" % name)
        for k, v_dict in m_attrs.items():
          if not type(k) == str:
            k = '.%s' % k.style.css_class.get_ref()
          style.append("  %s {%s; }" % (k, "; ".join(["%s: %s" % (i, ", ".join(j)) if isinstance(j, list) else "%s: %s" % (i, j) for i, j in v_dict.items()])))
        style.append("}")
    return "\n".join(style)

  def _repr_html_(self):
    """ Display for Jupyter """
    value = str(self)
    return value.replace(" ;", " ;\n  ").replace(" {", " {\n  ").replace("  }", "}")



"""
Common module for all the static data used in the different tests
"""
import os
import json
import webbrowser

# Deduce the \outs folder based on this module path
# the \outs folder is excluded from Git
TESTS_PATH = os.path.dirname(os.path.abspath(__file__))
OUTPUT_PATHS = r"%s\outs" % TESTS_PATH

# Extra static information with the links of the different external websites
URL_w3c = r"https://www.w3schools.com/html/tryit.asp?filename=tryhtml_basic"
URL_CODEPEN = r"https://codepen.io/"
URL_JSFIDDLE = r"https://jsfiddle.net/"

# Example files definition
MAP_COLS = {'flights.txt': {"distance": float, 'delay': float}}


def open_url(url):
  """
  Small utilities to open the browser to a specific web page.
  Some url are predefined and they are there to facilitate the testing of the generated Html pages

  :param url: A url

  """
  webbrowser.open(url)


def get_data(name, n=None, map_cols_type=None):
  """
  Load the static files used to test the different function.
  Files can be json or txt file (with header and delimited with a ,)

  :param name: THe filename
  :param n: Integer the max number of records
  :param map_cols_type: Dictionary. Special column type mapping

  :return: A record (A list of dictionaries)
  """
  if name.endswith(".json"):
    return json.load(open(os.path.join(TESTS_PATH, "data", name)))

  map_cols_type = map_cols_type or MAP_COLS.get(name, {})
  records = []
  with open(os.path.join(TESTS_PATH, "data", name)) as file:
    header = file.readline().strip().split(",")
    for i, line in enumerate(file):
      if n is not None and i > n:
        return records

      rec = dict(zip(header, line.strip().split(",")))
      for c, t in map_cols_type.items():
        rec[c] = t(rec[c])
      records.append(rec)
  return records




from epyk.core.js.html import JsHtml
from epyk.core.js.primitives import JsObjects
from epyk.core.js.fncs import JsFncs


class JsHtmlPanel(JsHtml.JsHtml):
  pass


class JsHtmlTr(JsHtml.JsHtml):
  display_value = "table-row"


class JsHtmlRow(JsHtml.JsHtml):
  display_value = "flex"


class JsHtmlGrid(JsHtml.JsHtml):

  @property
  def val(self):
    """

    :return:
    """
    return self._src.input.dom.val

  @property
  def content(self):
    return self._src.input.dom.content

  def panel(self, i):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param i:
    """
    panel = JsHtmlPanel(self._src, report=self._report)
    panel.varName = "%s.querySelector('.row').querySelector('div:nth-child(%s)')" % (self.varId, i+1)
    return panel

  @property
  def panels(self):
    for i, _ in enumerate(self._src.colsDim):
      yield self.panel(i)

    return ""

  def togglePanel(self, i):
    """
    Description:
    ------------
    Toggle the display of the column in a grid component.
    Teh other columns will be resized accordingly

    Attributes:
    ----------
    :param i: Integer. The column number (start at 0)

    :return:
    """
    return '''
      if(%(compId)s.querySelector('.row').querySelector('div:nth-child(%(i)s)').style.display == 'none'){
        %(compId)s.querySelector('.row').querySelector('div:nth-child(%(i)s)').style.display = 'block'
      } else {%(compId)s.querySelector('.row').querySelector('div:nth-child(%(i)s)').style.display = 'none'}
      ''' % {'compId': self.varId, 'i': i+1}


class JsHtmlTabs(JsHtml.JsHtml):

  def add_tab(self, name):
    return JsFncs.JsFunctions([
      JsObjects.JsNodeDom.JsDoms.new("div", varName="new_table"),
      JsObjects.JsNodeDom.JsDoms.new("div", varName="new_table_content"),
      JsObjects.JsNodeDom.JsDoms.get("new_table").css({"width": "100px", "display": 'inline-block', "vertical-align": "middle", "box-sizing": 'border-box',
                                                       'text-align': 'left'}),
      JsObjects.JsNodeDom.JsDoms.get("new_table_content").innerText(name),
      JsObjects.JsNodeDom.JsDoms.get("new_table_content").setAttribute("name", self._src.tabs_name),
      JsObjects.JsNodeDom.JsDoms.get("new_table_content").css({"width": "100px"}),
      JsObjects.JsNodeDom.JsDoms.get("new_table_content").css(self._src.options.css_tab),
      JsObjects.JsNodeDom.JsDoms.get("new_table_content").css({"padding": '5px 0'}),
      JsObjects.JsNodeDom.JsDoms.get("new_table").appendChild(JsObjects.JsObjects.get("new_table_content")),
      JsObjects.JsNodeDom.JsDoms.new("div", varName="tab_container"),
      self.querySelector("div").appendChild(JsObjects.JsObjects.get("new_table")),
     ])

  def tab(self, i):
    """
    Description:
    ------------
    Return the Javascript tab object

    Example
    tab.dom.tab(3).firstChild.css({"color": 'red'})

    :param i: Integer. Starting from 0 as we keep the Python indexing as reference
    :return:
    """
    return JsObjects.JsNodeDom.JsDoms.get("%s.firstChild.querySelector('div:nth-child(%s)')" % (self.varId, i+1))

  @property
  def selected_index(self):
    """
    Description:
    ------------
    Return the index of the selected tab

    :return: The index or -1
    """
    return JsObjects.JsObjects.get('''
      (function(node){ var selectedTab = node.querySelector('div[data-selected=true'); 
        if(selectedTab == null){ return -1; }
        else{ return selectedTab.getAttribute('data-index')}
      })(%s)''' % self.varId)

  @property
  def selected_name(self):
    """
    Description:
    ------------
    Return the name of the selected tab

    :return: The HTML content or an emtpy string
    """
    return JsObjects.JsObjects.get('''
          (function(node){ var selectedTab = node.querySelector('div[data-selected=true'); 
            if(selectedTab == null){ return ""; }
            else{ return selectedTab.innerHTML}
          })(%s)''' % self.varId)

  def deselect_tabs(self):
    """
    Description:
    ------------
    Deselect all the tabs in the component

    :return:
    """
    return JsFncs.JsFunctions([
      self._report.js.getElementsByName(self._src.tabs_name).all([
        self._report.js.data.all.element.setAttribute("data-selected", False),
        self._report.js.getElementsByName(self._src.tabs_name).all([
          self._report.js.data.all.element.css(self._src.options.tab_not_clicked_style())])
      ])
    ])



"""
Module in charge of connecting all the Microsoft Office tools

"""

import os

from epyk.core.js.Imports import requires


class DataOffice(object):
  class __internal(object):
    _props = {}

  def __init__(self, report=None):
    self._report = report if report is not None else self.__internal()

  def word(self, filename, path=None):
    """
    Open a word file

    This module will use an external Python package: python-docx

    Example
    docx = rptObj.data.office.word("FX Sales Trainee.docx", path=r"")
    print(docx.paragraphs)

    Related Pages:

			https://python-docx.readthedocs.io/en/latest/

    :param filename: The filename with the extension
    :param path: The path of the file
    :return: A docx Python object
    """
    docx = requires("docx", reason='Missing Package', install='python-docx', source_script=__file__, raise_except=True)

    doc = open(os.path.join(path, filename), 'rb')
    document = docx.Document(doc)
    doc.close()
    return document

  def ppt(self, filename, path=None):
    """
    Open a power point file

    Example
    ppts = rptObj.data.office.ppt("diagramme.pptx", r"")
    print(ppts.slides)

    Related Pages:

			https://python-pptx.readthedocs.io/en/latest/user/quickstart.html

    :param filename: The filename with the extension
    :param path: The path of the file

    :return: A pptx Python object
    """
    pptx = requires("pptx", reason='Missing Package', install='python-pptx', source_script=__file__, raise_except=True)

    return pptx.Presentation(os.path.join(path, filename))

  def xls(self, filename, path=None):
    """
    Open an excel file

    Example
    data = rptObj.data.office.xls("Classeur1.xlsx", "")
    print(data.sheet_by_index(0) )

    Related Pages:

			https://www.geeksforgeeks.org/reading-excel-file-using-python/

    :param filename: The filename with the extension
    :param path: The path of the file

    :return: A xlrd Python object
    """
    xlrd = requires("xlrd", reason='Missing Package', install='xlrd', source_script=__file__, raise_except=True)

    return xlrd.open_workbook(os.path.join(path, filename))

  def mdb(self, filename, path=None):
    """
    Open an Access file

    :param filename: The filename
    :param path: The database full path
    :rtype: epyk.core.py.Sql.SqlConnOdbc

    :return:
    """
    return self._report.data.db.mdb(filename, path)

  def outlook(self):
    """

    :return:
    """
    win32com = requires("win32com.client", reason='Missing Package', install='pywin32', source_script=__file__, raise_except=True)
    raise Exception("To be implemented")




from epyk.core import html


class Panels(object):
  def __init__(self, context):
    self.context = context

  def pills(self, color=None, width=(100, '%'), height=(None, 'px'), htmlCode=None, helper=None, options=None, profile=False):
    """
    Description:
    ------------
    Python wrapper to the Bootstrap Pills interface

    Usage::

      tab = rptObj.ui.panels.pills()
      for i in range(5):
        tab.add_panel("Panel %s" % i, rptObj.ui.text("test %s" % i))

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Tabs`

    Related Pages:

			https://getbootstrap.com/docs/4.0/components/navs/
    """
    dflt_options = {"css_tab": {'text-align': 'center', 'cursor': 'pointer', 'margin': '0 2px 0 0', 'border-radius': '5px',
                                'color': self.context.rptObj.theme.greys[-1], "background": self.context.rptObj.theme.greys[0]}}
    if options is not None:
      dflt_options.update(options)
    html_tabs = html.HtmlContainer.Tabs(self.context.rptObj, color, width, height, htmlCode, helper, dflt_options, profile)
    html_tabs.options.css_tab_clicked = {'color': html_tabs._report.theme.greys[0], 'background': html_tabs._report.theme.success[1]}
    self.context.register(html_tabs)
    return html_tabs

  def tabs(self, color=None, width=(100, '%'), height=(None, 'px'), htmlCode=None, helper=None, options=None, profile=False):
    """
    Description:
    ------------
    Python wrapper for a multi Tabs component

    Usage::

      tab = rptObj.ui.panels.tabs()
      for i in range(5):
        tab.add_panel("Panel %s" % i, rptObj.ui.text("test %s" % i))

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Tabs`

    Related Pages:

			https://getbootstrap.com/docs/4.0/components/navs/
    """
    dflt_options = {"css_tab": {'display': 'inline-block', 'text-align': 'center', 'cursor': 'pointer', 'margin': '0 2px 5px 0',
                                "border-bottom": "1px solid white"}}
    if options is not None:
      dflt_options.update(options)
    html_tabs = html.HtmlContainer.Tabs(self.context.rptObj, color, width, height, htmlCode, helper,
                                        dflt_options, profile)
    self.context.register(html_tabs)
    return html_tabs

  def arrows_up(self, color=None, width=(100, '%'), height=(None, 'px'), htmlCode=None, helper=None,
                options=None, profile=False):
    """
    Description:
    ------------
    Python wrapper for a multi Tabs component

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.TabsArrowsUp`

    Related Pages:

			https://getbootstrap.com/docs/4.0/components/navs/

    """
    dflt_options = {"css_tab": {'display': 'inline-block', 'text-align': 'center', 'cursor': 'pointer', 'margin': '0 2px 0 0',
                                "border-bottom": "1px solid white"}}
    if options is not None:
      dflt_options.update(options)
    html_tabs = html.HtmlContainer.TabsArrowsUp(self.context.rptObj, color, width, height, htmlCode, helper, dflt_options, profile)
    for t in html_tabs.tabs():
      t.style.add_classes.layout.panel_arrow_up()
    html_tabs.options.css_tab["color"] = html_tabs._report.theme.greys[-1]
    html_tabs.options.css_tab["height"] = "30px"
    html_tabs.options.css_tab_clicked = {"background": html_tabs._report.theme.success[1], "color": "white"}
    self.context.register(html_tabs)
    return html_tabs

  def arrows_down(self, color=None, width=(100, '%'), height=(None, 'px'), htmlCode=None, helper=None,
                  options=None, profile=False):
    """
    Description:
    ------------
    Python wrapper for a multi Tabs component

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.TabsArrowsDown`

    Related Pages:

			https://getbootstrap.com/docs/4.0/components/navs/
    """
    dflt_options = {
      "css_tab": {'display': 'inline-block', 'text-align': 'center', 'cursor': 'pointer', 'margin': '0 2px 0 0',
                 "border-bottom": "1px solid white"}}
    if options is not None:
      dflt_options.update(options)
    html_tabs = html.HtmlContainer.TabsArrowsDown(self.context.rptObj, color, width, height, htmlCode, helper, dflt_options, profile)
    for t in html_tabs.tabs():
      t.style.add_classes.layout.panel_arrow_down()
    html_tabs.options.css_tab["color"] = html_tabs._report.theme.greys[-1]
    html_tabs.options.css_tab["height"] = "30px"
    html_tabs.options.css_tab_clicked = {"background": html_tabs._report.theme.success[1], "color": "white"}
    self.context.register(html_tabs)
    return html_tabs

  def menu(self, color=None, width=(100, '%'), height=(None, 'px'), htmlCode=None, helper=None, options=None, profile=False):
    """
    Description:
    ------------
    Python wrapper to the Bootstrap Pills interface

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Tabs`

    Related Pages:

			https://getbootstrap.com/docs/4.0/components/navs/
    """
    dflt_options = {"css_tab": {'display': 'inline-block', 'text-align': 'center', 'cursor': 'pointer', 'margin': '0 2px 0 0',
                 'border-radius': '10px 10px 0 0'}}
    if options is not None:
      dflt_options.update(options)
    html_tabs = html.HtmlContainer.Tabs(self.context.rptObj, color, width, height, htmlCode, helper, dflt_options, profile)
    html_tabs.options.css_tab["color"] = html_tabs._report.theme.greys[-1]
    html_tabs.options.css_tab["background"] = html_tabs._report.theme.greys[0]
    html_tabs.options.css_tab_clicked = {'color': html_tabs._report.theme.greys[0], 'background': html_tabs._report.theme.success[1]}
    html_tabs.tabs_container.css({"border-bottom": "2px solid %s" % html_tabs._report.theme.success[1]})
    self.context.register(html_tabs)
    return html_tabs

  def sliding(self, htmlObjs, title, color=None, width=(100, "%"), height=(None, "px"),
            htmlCode=None, helper=None, options=None, profile=False):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.PanelSlide`

    Attributes:
    ----------
    :param htmlObjs:
    :param title:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    if htmlObjs is not None and not isinstance(htmlObjs, list):
      htmlObjs = [htmlObjs]
    html_slide = html.HtmlContainer.PanelSlide(self.context.rptObj, htmlObjs, title, color, width, height,
                                               htmlCode, helper, options or {}, profile)
    self.context.register(html_slide)
    return html_slide

  def split(self, left=None, right=None, width=(100, '%'), height=(200, 'px'), left_width=(160, 'px'), resizable=True,
            helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      number = rptObj.ui.rich.number(500, "Test", height=(150, 'px'))
      number_2 = rptObj.ui.rich.number(500, "Test 2 ", options={"url": "http://www.google.fr"})
      div = rptObj.ui.layouts.panelsplit(left=number, right=number_2)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.PanelSlide`

    Related Pages:

			https://codepen.io/rstrahl/pen/eJZQej

    Attributes:
    ----------
    :param width:
    :param height:
    :param left_width:
    :param left:
    :param right:
    :param resizable:
    :param helper:
    :param profile:
    """
    html_split = html.HtmlContainer.PanelSplit(self.context.rptObj, width, height, left_width, left, right, resizable, helper, profile)
    self.context.register(html_split)
    return html_split




from epyk.core.data import DataClass


class KeyTable(DataClass):

  def activate(self):
    self.columns = ":not(:first-child)"
    return self

  @property
  def blurable(self):
    """
    Allow KeyTable's focus to be blurred (removed) from a table.

    https://datatables.net/reference/option/keys.blurable
    """
    return self._attrs["blurable"]

  @blurable.setter
  def blurable(self, val):
    self._attrs["blurable"] = val

  @property
  def className(self):
    """
    Set the class name used for the focused cell.

    https://datatables.net/reference/option/keys.className
    """
    return self._attrs["className"]

  @className.setter
  def className(self, val):
    self._attrs["className"] = val

  @property
  def clipboard(self):
    """
    Enable / disable clipboard interaction with KeyTable.

    https://datatables.net/reference/option/keys.clipboard
    """
    return self._attrs["clipboard"]

  @clipboard.setter
  def clipboard(self, val):
    self._attrs["clipboard"] = val

  @property
  def clipboardOrthogonal(self):
    """
    Set the orthogonal data to copy to clipboard.

    https://datatables.net/reference/option/keys.clipboardOrthogonal
    """
    return self._attrs["clipboardOrthogonal"]

  @clipboardOrthogonal.setter
  def clipboardOrthogonal(self, val):
    self._attrs["clipboardOrthogonal"] = val

  @property
  def columns(self):
    """
    Select the columns that can gain focus.

    https://datatables.net/reference/option/keys.columns
    """
    return self._attrs["columns"]

  @columns.setter
  def columns(self, val):
    self._attrs["columns"] = val

  @property
  def focus(self):
    """
    Cell to receive initial focus in the table.

    https://datatables.net/reference/option/keys.focus
    """
    return self._attrs["focus"]

  @focus.setter
  def focus(self, val):
    self._attrs["focus"] = val

  @property
  def keys(self):
    """
    Limit the keys that KeyTable will listen for and take action on

    https://datatables.net/reference/option/keys.keys
    """
    return self._attrs["keys"]

  @keys.setter
  def keys(self, val):
    self._attrs["keys"] = val

  @property
  def tabIndex(self):
    """
    Set the table's tab index for when it will receive focus.

    https://datatables.net/reference/option/keys.tabIndex
    """
    return self._attrs["tabIndex"]

  @tabIndex.setter
  def tabIndex(self, val):
    self._attrs["tabIndex"] = val



"""

"""

from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesChart


class CatalogChart(Catalog.CatalogGroup):

  def div(self):
    """  """
    return self._set_class(CssStylesChart.CssDivChart)

  def billboard_title(self):
    """  """
    return self._set_class(CssStylesChart.CssBillboardTitle)

  def billboard_legend(self):
    """  """
    return self._set_class(CssStylesChart.CssBillboardLegend)

  def billboard_axis(self):
    """  """
    return self._set_class(CssStylesChart.CssBillboardAxis)

  def billboard_axis_x(self):
    """  """
    return self._set_class(CssStylesChart.CssBillboardXAxis)

  def billboard_axis_y(self):
    """  """
    return self._set_class(CssStylesChart.CssBillboardYAxis)

  def c3_title(self):
    """  """
    return self._set_class(CssStylesChart.CssC3Title)

  def c3_legend(self):
    """  """
    return self._set_class(CssStylesChart.CssC3Legend)

  def c3_axis(self):
    """  """
    return self._set_class(CssStylesChart.CssC3Axis)

  def c3_axis_x(self):
    """  """
    return self._set_class(CssStylesChart.CssC3XAxis)

  def c3_axis_y(self):
    """  """
    return self._set_class(CssStylesChart.CssC3YAxis)

  def nvd3_axis(self):
    """  """
    return self._set_class(CssStylesChart.CssNVD3Axis)

  def nvd3_axis_label(self):
    """  """
    return self._set_class(CssStylesChart.CssNVD3AxisLabel)

  def nvd3_axis_legend(self):
    """  """
    return self._set_class(CssStylesChart.CssNVD3AxisLegend)

  def nvd3_grid_hide(self):
    """  """
    return self._set_class(CssStylesChart.CssNVD3HideGrid)

  def sparklines(self):
    """  """
    return self._set_class(CssStylesChart.CssSparklines)



from epyk.core.js import JsUtils


class JsIf(object):

  def __init__(self, jsCondition, jsFncs, context=None):
    """
    Description:
    ------------
    Create a JavaScript If statement

    Usage::

      JsIf.JsIf(self.input.dom.hasClass("fa-check"), jsFncsTrue)

    Attributes:
    ----------
    :param jsCondition: The Javascript condition. Can be a JsBoolean object
    :param jsFncs: The Javascript functions
    :param context: Optional. Dictionary. Meta data concerning the context
    """
    self._context = context
    jsFncs = JsUtils.jsConvertFncs(jsFncs, False)
    self._js = [(jsCondition, jsFncs)]
    self.__jsElse = None

  def elif_(self, jsCondition, jsFncs):
    """
    Description:
    ------------
    Add a Javascript elif statement to the loop

    Usage::

      Attributes:
    ----------
    :param jsCondition: The Javascript condition. Can be a JsBoolean object
    :param jsFncs: The Javascript functions
    :return: The If object to allow the chaining
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs, False)
    self._js.append((jsCondition, jsFncs))
    return self

  def else_(self, jsFncs):
    """
    Description:
    ------------
    Add the Javascript else statement to the loop

    Usage::

      JsIf.JsIf(self.input.dom.hasClass("fa-check"), jsFncsTrue).else_(jsFncFalse)

    Attributes:
    ----------
    :param jsFncs: The Javascript functions
    :return: The If object to allow the chaining
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs, False)
    self.__jsElse = jsFncs
    return self

  def toStr(self):
    strData = ["if(%s){%s}" % (self._js[0][0], ";".join(map(lambda x: str(x),  self._js[0][1])))]
    for condition, fncs in self._js[1:]:
      strData.append("else if(%s){%s}" % (condition, ";".join(fncs)))
    if self.__jsElse is not None:
      strData.append("else{%s}" % ";".join(map(lambda x: str(x),  self.__jsElse)))
    self._js, self.__jsElse = [], None # empty the stack
    return "".join(strData)



"""
Common module for managing dates.

This module is a light wrapper on top of datetime in order to perform basic operations on dates.
This will also standardise the date format to YYYY-MM-DD in the Python layer to simplify the
conversion to the Javascript

All the tests in this module are using doctest
Related Pages:

		https://docs.python.org/2/library/doctest.html
"""

import time
import datetime


class PyDates(object):
  class __internal(object):
    _props, _context, jsOnLoadEvtsFnc = {}, {}, []

  def __init__(self, src=None):
    self.__src = src if src else self.__internal()

  @property
  def today(self):
    """
    Return a String date in a format YYYY-MM-DD

    Even if within the property python date object are used, this function will
    always return a string date in a specific format to guarantee and simplify the compatibility between languages
    within the components

    Example
    PyDates().today

    Related Pages:

			https://docs.python.org/2/library/datetime.html

    :return: A string date in the format YYYY-MM-DD
    """
    return datetime.datetime.today().strftime('%Y-%m-%d')

  @property
  def now(self):
    """
    Return the current timestamp in a format YYYY-MM-DD HH:mm:dd

    Example
    PyDates().now

    Related Pages:

			https://docs.python.org/2/library/datetime.html

    :return: Return a string timestamp
    """
    return datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')

  @property
  def cob(self):
    """
    Returns the last close of business date

    In this property the parameter weekdays is forced to True.

    Example
    rptObj.py.dates.cob

    :return: A string date
    """
    return self.date_from_alias("T")

  @property
  def month_end(self):
    """
    Returns the last month end date

    In this property the parameter weekdays is forced to True.

    Example
    rptObj.py.dates.month_end

    :return: A string date
    """
    return self.date_from_alias("M")

  @property
  def months(self):
    """
    Returns the list of month end dates from the beginning of the year

    In this property the parameter weekdays is forced to True.

    Example
    rptObj.py.dates.months

    :return: A list of String dates
    """
    date_split = list(map(lambda x: int(x), self.today.split("-")))
    return self.month_ends(datetime.datetime(date_split[0], 1, 1).strftime('%Y-%m-%d'), self.today)

  @property
  def quarters(self):
    """
    Return the list of quarter dates since the beginning of the year

    In this property the parameter weekdays is forced to True.

    Example
    rptObj.py.dates.quarters

    :return: A list of String dates
    """
    results = []
    for i, v in enumerate(self.months):
      if i % 3 == 2:
        results.append(v)
    return results

  def date_from_alias(self, alias, from_date=None):
    """
    Return the date corresponding to an alias code like T, T-N, M...

    Example
    >>> PyDates().date_from_alias("T", "2019-08-08")
    '2019-08-07'

    :param alias: The alias of the operation (T-3, M-2....)
    :param from_date: The start date from which the time operation is applied. Today by default

    :return: The converted date or a list of dates
    """
    if from_date is None:
      cob_date = datetime.datetime.today()
    else:
      cob_date = datetime.datetime(*map(lambda x: int(x), from_date.split("-")))
    if len(alias) > 1:
      f_type, f_count = alias[0], "".join(alias[2:])
    else:
      f_type, f_count = alias, 0
    if f_type == 'T':
      for i in range(0, int(f_count) + 1):
        if len(alias) > 1:
          if alias[1] == '+':
            cob_date = cob_date + datetime.timedelta(days=1)
            while cob_date.weekday() in [5, 6]:
              cob_date = cob_date + datetime.timedelta(days=1)
          else:
            cob_date = cob_date - datetime.timedelta(days=1)
            while cob_date.weekday() in [5, 6]:
              cob_date = cob_date - datetime.timedelta(days=1)
        else:
          cob_date = cob_date - datetime.timedelta(days=1)
          while cob_date.weekday() in [5, 6]:
            cob_date = cob_date - datetime.timedelta(days=1)
      return cob_date.strftime('%Y-%m-%d')

    if f_type == 'M':
      end_month_date = datetime.datetime(cob_date.year, cob_date.month - int(f_count), 1)
      end_month_date = end_month_date - datetime.timedelta(days=1)
      while end_month_date.weekday() in [5, 6]:
        end_month_date = end_month_date - datetime.timedelta(days=1)
      return end_month_date.strftime('%Y-%m-%d')

    if f_type == 'W':
      cob_date = cob_date - datetime.timedelta(days=1)
      while cob_date.weekday() != 4:
        cob_date = cob_date - datetime.timedelta(days=1)
      cob_date = cob_date - datetime.timedelta(days=(int(f_count) * 7))
      return cob_date.strftime('%Y-%m-%d')

    if f_type == 'Y':
      end_year_date = datetime.datetime(cob_date.year - int(f_count), 1, 1)
      end_year_date = end_year_date - datetime.timedelta(days=1)
      while end_year_date.weekday() in [5, 6]:
        end_year_date = end_year_date - datetime.timedelta(days=1)
      return end_year_date.strftime('%Y-%m-%d')

    return alias

  def date_from_excel(self, xlDate):
    """
    Convert a Excel date to a AReS standard date format YYYY-MM-DD.

    Example
    >>> PyDates().date_from_excel(39448)
    '2008-01-01'

    Related Pages:

			https://support.office.com/en-gb/article/date-function-e36c0c8c-4104-49da-ab83-82328b832349?ui=en-US&rs=en-GB&ad=GB

    :param xlDate:

    :return: The date as a String in the common format YYYY-MM-DD in AReS
    """
    dt = datetime.datetime.fromordinal(datetime.datetime(1900, 1, 1).toordinal() + xlDate - 2)
    return dt.strftime('%Y-%m-%d')

  def month_ends(self, from_dt, to_dt, weekdays=True):
    """
    Return the list of end of month dates between two dates

    Example
    >>> PyDates().month_ends("2019-01-01", "2019-06-05", False)
    ['2019-01-31', '2019-02-28', '2019-03-31', '2019-04-30', '2019-05-31']

    :param from_dt: The start date in format YYYY-MM-DD
    :param to_dt: The end date in format YYYY-MM-DD
    :param weekdays: remove the weekends from the potential dates (take the day before). Default True

    :return: A list of dates.
    """
    if to_dt < from_dt:
      tmp = from_dt
      from_dt = to_dt
      to_dt = tmp
    results = []
    date_split = list(map(lambda x: int(x), from_dt.split("-")))
    end_dt = datetime.datetime(*map(lambda x: int(x), to_dt.split("-")))
    dt = datetime.datetime(date_split[0], date_split[1]+1, 1) - datetime.timedelta(days=1)
    while dt < end_dt:
      results.append(dt.strftime('%Y-%m-%d'))
      dt = datetime.datetime(dt.year + int((dt.month + 1) / 12), (dt.month + 1) % 12 + 1, 1)
      dt = dt - datetime.timedelta(days=1)
      if weekdays:
        while dt.weekday() in [5, 6]:
          dt = dt - datetime.timedelta(days=1)
    return results

  def range_dates(self, from_dt, to_dt, weekdays=True):
    """
    Get the list of dates between two dates.

    The date should be two string dates in the format YYYY-MM-DD.
    The resulting range of date will always be increasing

    Example
    >>> PyDates().range_dates("2019-01-01", "2019-01-11")
    ['2019-01-11', '2019-01-10', '2019-01-09', '2019-01-08', '2019-01-07', '2019-01-04', '2019-01-03', '2019-01-02', '2019-01-01']

    :param from_dt: The start date in format YYYY-MM-DD
    :param to_dt: The end date in format YYYY-MM-DD
    :param weekdays: remove the weekends from the potential dates (take the day before). Default True

    :return: A list of dates.
    """
    start_date = self.date_from_alias(from_dt)
    end_date = self.date_from_alias(to_dt, from_date=start_date)
    if start_date < end_date:
      start_date = self.date_from_alias(to_dt, from_date=start_date)
      end_date = self.date_from_alias(from_dt)
    dt = datetime.datetime(*map(lambda x: int(x), start_date.split('-')))
    target_date = datetime.datetime(*map(lambda x: int(x), end_date.split('-')))
    dates = [start_date]
    while dt > target_date:
      dt = dt - datetime.timedelta(days=1)
      if not dt.weekday() in [5, 6] or not weekdays:
        dates.append(dt.strftime('%Y-%m-%d'))
    return dates

  def from_timestamp(self, timestamp, offset=0, reference=60, factor=1000):
    """
    The default value will be given considering the GMT time

    Example
    timestamp_s = rptObj.py.dates.from_timestamp(1573074335010, 0)

    :param timestamp: Integer. The timestamp in miliseconds
    :param offset: Integer. The time zone
    :param reference: Integer. The reference shift in minutes

    :return: The server timestamp string
    """
    date = datetime.datetime.fromtimestamp(timestamp / factor)
    date = date + datetime.timedelta(minutes=(int(offset) + reference))
    return date.strftime('%Y-%m-%d %H:%M:%S')

  def to_server_time(self, timestamp, offset=0, reference=60):
    """
    Return the converted timestamp to be stored in the database.
    This conversion will be based on the offset coming from the UI to convert to common time

    Example
    >>> PyDates().to_server_time("2019-08-20 20:04:10", 2)
    '2019-08-20 21:06:10'

    :param timestamp: The client timestamp
    :param offset: The client offset time to be applied before storage in hour
    :param reference: The reference time used on the server side

    :return: The server timestamp string
    """
    date = datetime.datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
    date = date + datetime.timedelta(minutes=(int(offset) + reference))
    return date.strftime('%Y-%m-%d %H:%M:%S')

  def to_user_time(self, timestamp, offset, reference=60):
    """
    Return the converted timestamp to be returned to the user.
    This is converting a stored timestamp to a user one.

    Example
    >>> PyDates().to_user_time('2019-08-20 21:06:10', 2)
    '2019-08-20 20:04:10'

    :param timestamp: The server timestamp
    :param offset: The client offset time to be applied before storage
    :param reference: The reference time used on the server side

    :return: The client timestamp string
    """
    date = datetime.datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
    date = date + datetime.timedelta(minutes=-1 * (int(offset) + reference))
    return date.strftime('%Y-%m-%d %H:%M:%S')




from epyk.core.css.styles import GrpCls
from epyk.core.css import Classes


class ClassSelect(GrpCls.ClassHtml):

  def __init__(self, htmlObj):
    super(ClassSelect, self).__init__(htmlObj)
    self._css_class_dt, self._css_class_dt_ui, self._css_select = None, None, None
    self._css_select_input, self._css_item_option, self._css_item_selected = None, None, None
    self._css_menu_li, self._css_select_search, self._css_select_menu_hover = None, None, None
    self.classList['main'].add(self.cls_select)
    self.classList['main'].add(self.cls_select_button)
    self.classList['main'].add(self.cls_select_outline)
    self.classList['other'].add(self.cls_select_input)
    self.classList['main'].add(self.cls_item_option)
    self.classList['other'].add(self.cls_item_selected)
    self.classList['other'].add(self.cls_toggle)

  @property
  def cls_item_selected(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_item_selected is None:
      self._css_item_selected = Classes.CatalogSelect.CatalogSelect(self.htmlObj._report, self.classList['other']).selected()
    return self._css_item_selected

  @property
  def cls_toggle(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_menu_li is None:
      self._css_menu_li = Classes.CatalogSelect.CatalogSelect(self.htmlObj._report, self.classList['other']).toggle()
    return self._css_menu_li

  @property
  def cls_select(self):
    """

    :rtype: Classes.CatalogInput.CatalogDate
    """
    if self._css_select is None:
      self._css_select = Classes.CatalogSelect.CatalogSelect(self.htmlObj._report, self.classList['main']).base()
    return self._css_select

  @property
  def cls_select_button(self):
    """

    :rtype: Classes.CatalogInput.CatalogDate
    """
    if self._css_class_dt is None:
      self._css_class_dt = Classes.CatalogSelect.CatalogSelect(self.htmlObj._report, self.classList['main']).button()
    return self._css_class_dt

  @property
  def cls_select_outline(self):
    """

    :rtype: Classes.CatalogInput.CatalogDate
    """
    if self._css_class_dt_ui is None:
      self._css_class_dt_ui = Classes.CatalogSelect.CatalogSelect(self.htmlObj._report, self.classList['main']).outline()
    return self._css_class_dt_ui

  @property
  def cls_item_option(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_item_option is None:
      self._css_item_option = Classes.CatalogSelect.CatalogSelect(self.htmlObj._report, self.classList['main']).item()
    return self._css_item_option

  @property
  def cls_select_input(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_select_input is None:
      self._css_select_input = Classes.CatalogSelect.CatalogSelect(self.htmlObj._report, self.classList['other']).search_box_input()
    return self._css_select_input


class ClassDropDown(GrpCls.ClassHtml):

  def __init__(self, htmlObj):
    super(ClassDropDown, self).__init__(htmlObj)
    self._css_base, self._css_menu, self._css_menu_after, self._css_menu_link = None, None, None, None
    self._css_menu_hover, self._css_menu_pull_left, self._css_menu_li = None, None, None
    self.classList['main'].add(self.cls_base)
    self.classList['main'].add(self.cls_menu)
    self.classList['main'].add(self.cls_menu_li)
    self.classList['main'].add(self.cls_menu_after)
    self.classList['main'].add(self.cls_menu_hover)
    self.classList['main'].add(self.cls_menu_pull_left)

  @property
  def cls_base(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_base is None:
      self._css_base = Classes.CatalogTree.CatalogDropDown(self.htmlObj._report, self.classList['main']).base()
    return self._css_base

  @property
  def cls_menu(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_menu is None:
      self._css_menu = Classes.CatalogTree.CatalogDropDown(self.htmlObj._report, self.classList['main']).menu()
    return self._css_menu

  @property
  def cls_menu_li(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_menu_li is None:
      self._css_menu_li = Classes.CatalogTree.CatalogDropDown(self.htmlObj._report, self.classList['main']).menu_li()
    return self._css_menu_li

  @property
  def cls_menu_after(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_menu_after is None:
      self._css_menu_after = Classes.CatalogTree.CatalogDropDown(self.htmlObj._report, self.classList['main']).menu_after()
    return self._css_menu_after

  @property
  def cls_menu_hover(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_menu_hover is None:
      self._css_menu_hover = Classes.CatalogTree.CatalogDropDown(self.htmlObj._report, self.classList['main']).menu_hover()
    return self._css_menu_hover

  @property
  def cls_menu_pull_left(self):
    """

    :rtype: Classes.CatalogTree.CatalogDropDown
    """
    if self._css_menu_pull_left is None:
      self._css_menu_pull_left = Classes.CatalogTree.CatalogDropDown(self.htmlObj._report, self.classList['main']).menu_pull_left()
    return self._css_menu_pull_left




import os
import re
import json
import functools

from epyk.core.js import Imports
from epyk.core.js.primitives import JsObject


# --------------------------------------------------------------------------------------------------------------
#                                                       DECORATORS
#
def incompatibleBrowser(browsers):
  """
  Description:
  ------------

  :param browsers: List. Incompatible browsers
  """
  def decorator(func):
    @functools.wraps(func)
    def decorated(*args, **kwargs):
      print("Warning: This function - %s - is not compatible with %s" % (func.__name__, ", ".join(browsers)))
      return func(*args, **kwargs)

    return decorated

  return decorator


def fromVersion(data):
  """
  Description:
  ------------
  This system decorate will decorate a component function to specify during the Python execution
  if a method is not yet available in the current state of the Javascript modules.

  Example
  .fromVersion({'jqueryui': '1.12.0'})

  :param data: Set the minimum version of a package for a specific function

  :return: The decorated function
  """
  def decorator(func):
    @functools.wraps(func)
    def decorated(*args, **kwargs):
      for k, v in data.items():
        if k in Imports.JS_IMPORTS:
          for mod in Imports.JS_IMPORTS[k]['modules']:
            if mod['version'] < v:
              raise Exception("Function %s can only be used from %s version %s (current %s)" % (func.__name__, k, v, mod['version']))

      return func(*args, **kwargs)

    return decorated

  return decorator


def untilVersion(data, newFeature):
  """
  Description:
  ------------
  This system decorate will decorate a component function to specify during the Python execution
  if a method is not available since the current state of the Javascript modules.
  Indeed as it is possible to override JS version on the fly it is also important to get notify with functions
  are not compatible anymore.
  This decorator will also propose an alternative with the new feature to.

  Usage::

      .untilVersion({'jqueryui': '1.12.0'}, "new function")

  :param data: String. The maximum version number of a package to use this function
  :param newFeature: String. The new function name

  :return: The decorated function
  """
  def decorator(func):
    @functools.wraps(func)
    def decorated(*args, **kwargs):
      for k, v in data.items():
        if k in Imports.JS_IMPORTS:
          for mod in Imports.JS_IMPORTS[k]['modules']:
            if mod['version'] > v:
              raise Exception("Function %s can only be used since %s version %s (current %s). It has been replaced by %s" % (func.__name__, k, v, mod['version'], newFeature))

      return func(*args, **kwargs)
    return decorated
  return decorator


# --------------------------------------------------------------------------------------------------------------
#                                                       FUNCTIONS
#
def jsConvertData(jsData, jsFnc):
  """
  Description:
  ------------
  Generic conversion function for any data in the internal framework.
  This will convert to String any data coming from the Javascript Python interface.

  Any pure Python object will be converted using the json function to be then written as a string
  to the resulting page

  :param jsData: The Python Javascript data
  :param jsFnc: Optional. The conversion function

  :return:
  """
  if not hasattr(jsData, 'varData') and not hasattr(jsData, 'fncName'):
    if hasattr(jsData, 'toStr'):
      return jsData.toStr()
    else:
      try:
        return JsObject.JsObject(json.dumps(jsData))

      except TypeError as err:
        return str(jsData)

      except Exception as err:
        if isinstance(jsData, range):
          return JsObject.JsObject(json.dumps(list(jsData)))

        raise

  return jsData


def jsConvert(jsData, jsDataKey, isPyData, jsFnc):
  """
  Description:
  ------------

  :param jsData:
  :param jsDataKey:
  :param isPyData:
  :param jsFnc:
  """
  if isPyData:
    if hasattr(jsData, 'toStr'):
      return jsData.toStr()
    else:
      try:
        return json.dumps(jsData)

      except Exception as err:
        if isinstance(jsData, range):
          return json.dumps(list(jsData))

        raise

  if jsDataKey is not None:
    jsData = "%s['%s']" % (jsData, jsDataKey)
  if jsFnc is not None:
    jsData = "%s(%s)" % (jsFnc, jsData)
  return jsData


def getJsValid(value, fail=True):
  """
  Return an error if the variable name is not valid following the Javascript naming conventions.
  Even if the function will fail it will propose a valid name to replace the one passed in input

  Example
  >>> getJsValid("test-js", False)
  'testjs'

  >>> getJsValid("234@test-js", False)
  'js234testjs'

  >>> getJsValid("234@test-js", True)
  Traceback (most recent call last):
      ...
  Exception: Javascript Variable name 234@test-js, for example you could use js234testjs instead

  Related Pages:

			https://www.w3schools.com/js/js_conventions.asp

  :param value: The Javascript variable name
  :param fail: Boolean to raise an exception if the name is not valid on the Javascript side

  :return: The input variable name or a suggested one
  """
  regex = re.compile('[^a-zA-Z0-9_]')
  cleanName = regex.sub('', value.strip())
  isValid = not value[0].isdigit() and cleanName == value
  if fail and not isValid:
    raise Exception("Javascript Variable name %s, for example you could use js%s instead" % (value, cleanName))

  if cleanName[0].isdigit():
    cleanName = "js%s" % cleanName
  return cleanName


def jsConvertFncs(jsFncs, isPyData=False, jsFncVal=None, toStr=False):
  """
  Description:
  ------------
  Generic conversion function for all the PyJs functions

  :param jsFncs: Array. The PyJs functions
  :param isPyData: Boolean. A flag to force the Python conversion using json
  :param jsFncVal:
  :param toStr: Boolean. A flag to specify if the result should be aggregated

  :return:
  """
  if not isinstance(jsFncs, list):
    jsFncs = [jsFncs]

  cnvFncs = []
  for f in jsFncs:
    if f is None:
      continue

    if hasattr(f, 'toStr'):
      strFnc = f.toStr()
      if jsFncVal is not None:
        strFnc = strFnc.replace("trans_val", jsFncVal)
      cnvFncs.append(strFnc)
    else:
      if isPyData:
        strVal = json.dumps(f)
        if jsFncVal is not None:
          strVal = strVal.replace("trans_val", jsFncVal)
        cnvFncs.append(strVal)
      else:
        if jsFncVal is not None:
          f = str(f).replace("trans_val", jsFncVal)
        cnvFncs.append(f)
  if toStr:
    return ";".join(cnvFncs)

  return cnvFncs


def cleanFncs(fnc):
  """
  Description:
  ------------
  Try to remove as much as possible all the characters in order to speed up the javascript
  Indeed most of the browsers are using minify Javascript to make the page less heavy

  Thus pre stored function code can be written to be easier to read.

  :param fnc: The Javascript String

  :return: Return a cleaned an minify Javascript String
  """
  return "".join([r.strip() for r in fnc.strip().split('\n')])


def isNotDefined(varName):
  """
  Description:
  ------------
  Check if a variable is defined

  Example
  JsUtils.isNotDefined(varId)

  :param varName: String. The varName
  :return: A string in Python and a Boolean in Javascript
  """
  return "typeof %s === 'undefined'" % varName


class JsFile(object):
  def __init__(self, scriptName=None, path=None):
    self.scriptName, self.path = scriptName, path
    self.file_path = os.path.join(path, "js") # all the files will be put in a common directory
    if not os.path.exists(self.file_path):
      os.mkdir(self.file_path)
    self.__data = []

  def writeJs(self, jsFncs):
    """
    Write the Javascript piece of code to the file

    Example
    dt = JsDate.new("2019-05-03")
    f.writeJs([dt,
      Js.JsConsole().log(dt.getDay()),
      Js.JsConsole().log(dt.getFullYear())])

    :param jsFncs: The Javascript fragments

    :return: The File object
    """
    self.__data.extend(jsConvertFncs(jsFncs))
    return self

  def writeReport(self, rptObj):
    """
    Write the Javascript content of a report to a structure .js file.
    This could help on the investigation and can be directly used in Codepen for testing

    :param rptObj: The report object

    :return:
    """
    props = rptObj._src._props if hasattr(rptObj, '_src') else rptObj._props
    for data_id, data in props.get("data", {}).get('sources', {}).items():
      self.__data.append("var data_%s = %s" % (data_id, json.dumps(data)))

    for k, v in props.get('js', {}).get('functions', {}).items():
      sPmt = "(%s)" % ", ".join(list(v["pmt"])) if "pmt" in v else "{}"
      self.__data.append("function %s%s{%s}" % (k, sPmt, v["content"].strip()))

    for c, d in props.get('js', {}).get("constructors", {}).items():
      self.__data.append(d)

    for c, d in props.get('js', {}).get("datasets", {}).items():
      self.__data.append(d)

    for b in props.get('js', {}).get("builders", []):
      self.__data.append(b)

    keyboardShortcuts = props.get('js', {}).get('keyboard', {})
    if keyboardShortcuts:
      self.__data.append("document.addEventListener('keydown', function(e){var code = e.keyCode || e.which")
      for k, v in keyboardShortcuts.items():
        self.__data.append("if(%s){%s}" % (k, v))
      self.__data.append("})")

  def codepen(self, rptObj, cssObj=None, target='_self'):
    """
    Send the piece of Javascript to Codepen for testing

    Related Pages:

			https://codepen.io/

    :param rptObj: The report object
    :param cssObj: The internal CSS object from the page
    :param target: A string flag to specify the target page in the browser
    """
    import webbrowser

    if hasattr(rptObj, '_to_html_obj'):
      results = rptObj._to_html_obj(content_only=True)
      js_external = re.findall('<script language="javascript" type="text/javascript" src="(.*?)"></script>', results['jsImports'])
      result = {"js": results["jsFrgs"], "js_external": ";".join(js_external)}
    else:
      self.writeReport(rptObj)
      import_obj = Imports.ImportManager(online=True)
      css_external = import_obj.cssURLs(import_obj.cssResolve(rptObj._src.cssImport))
      js_external = import_obj.jsURLs(import_obj.jsResolve(rptObj._src.jsImports))
      result = {"js": ";".join(self.__data), "js_external": ";".join(js_external), "css_external": ";".join(css_external)}
    if cssObj is not None:
      result["css"] = cssObj.toCss()
    data = rptObj.location.postTo("https://codepen.io/pen/define/", {"data": json.dumps(result)}, target=target)
    outFile = open(os.path.join(self.file_path, "CodePenJsLauncher.html"), "w")
    outFile.write('<html><body></body><script>%s</script></html>' % data.replace("\\\\n", ""))
    webbrowser.open(outFile.name)

  def close(self, jsObj=None):
    """
    Write the file and close the buffer

    :param jsObj: The internal JsObject
    """
    src_obj = jsObj._src if hasattr(jsObj, '_src') else jsObj
    outFile = open(os.path.join(self.file_path, "%s.js" % self.scriptName), "w")
    js_external = ""
    if jsObj is not None:
      outFile.write("//Javascript Prototype extensions \n\n")
      for fnc, details in src_obj._props.get('js', {}).get('prototypes', {}).items():
        outFile.write("%s = function(%s){%s};" % (fnc, ",".join(details.get('pmts', [])), details["content"]))
      outFile.write("\n\n//Javascript Global functions \n\n")
      for fnc, details in src_obj._props.get('js', {}).get('functions', {}).items():
        outFile.write("function %s(%s){%s}" % (fnc, ",".join(details.get('pmt', [])), details["content"]))
      import_obj = Imports.ImportManager(online=True)
      js_external = import_obj.jsResolve(src_obj.jsImports)
    outFile.write("\n\n")
    outFile.write("//Javascript Data\n\n")
    for data_id, data in src_obj._props.get("data", {}).get('sources', {}).items():
      outFile.write("var data_%s = %s" % (data_id, json.dumps(data)))
    outFile.write("\n\n")
    outFile.write("//Javascript functions\n\n")
    if len(src_obj._props.get('js', {}).get('onReady', [])) > 0:
      outFile.write("%s;" % jsConvertFncs(src_obj._props.get('js', {}).get('onReady', []), toStr=True))
    outFile.write(";".join(self.__data))
    outFile.close()
    with open(r"%s\Launche_%s.html" % (self.path, self.scriptName), "w") as f:
      f.write('<html><head></head><body></body>%s<script src="js/%s.js"></script></html>' % (js_external, self.scriptName))
    return r"%s\Launche_%s.html" % (self.path, self.scriptName)



"""

"""


import sys
import inspect

factory = None


def getAggFnc():
  """
  Load all the aggregation definitions defined in this module in a factory.

  This will be structured as a dictionary with as key the name of the aggregator and as value the object.
  This factory will store directly the object as the structure cannot be dynamic and it is not expecting any input.
  The dynamic part will be added as input of the toJs() function.
  """
  global factory

  if factory is None:
    tmpFactory = {}
    for name, obj in inspect.getmembers(sys.modules[__name__], inspect.isclass):
      if getattr(obj, 'name', None) not in [None, '__main__'] and issubclass(obj, JsPivotAggFnc):
        tmpFactory[obj.name] = obj()
    # Atomic function to avoid asynchronous clashes on the server
    factory = tmpFactory
  return factory


class JsPivotAggFnc(object):
  """
  Based abstract class for the Javascript Pivot aggregator.

  The base class will define the slots of the expected class variables that the child classes should defined.
  The toJs function will, thanks to the options inputs, will create the Javascript dictionary that the PivotTable
  library is expecting.
  """
  name = None
  __slots__ = ['keyAgg', 'key2Agg', 'numInputs', 'push', 'value', 'format']

  def toJs(self, options):
    """
    Convert the aggregator object to a Javascript dictionary usable in the Pivot Table javascript library.
    All the parameters will be standard to the Js module and this will convert the Python object to the
    corresponding Javascript ones.
    """
    jsPivot = ["tmpVal: 0"]
    jsMapFncs = {'push': "function(record) {%s}", 'value': "function() {%s}", 'format': 'function(x) {%s}'}
    _opts = dict(getattr(self, '_dflts', {}))
    _opts.update(options)
    self.numInputs = 2 if getattr(self, "key2Agg", None) is not None else 1
    for slot in self.__slots__:
      slotVal = getattr(self, slot) % _opts if slot in ['push', 'value', 'format'] else getattr(self, slot)
      if slotVal is None:
        continue

      if slot in jsMapFncs:
        if slot in ['value', 'format'] and not 'return ' in slotVal:
          raise Exception("value and format should return a value")

        jsPivot.append("%s: %s" % (slot, jsMapFncs[slot] % slotVal))
      else:
        jsPivot.append("%s: %s" % (slot, slotVal))
    return "{%s}" % ", ".join(jsPivot)


# -------------------------------------------------------------------------------------------------------------------
#                           SIMPLE TRANSFORMATION USING ONE KEY
#
class JsPivotSumAgg(JsPivotAggFnc):
  """
  Aggregator in charge of adding up all the values based on the defined rows.
  This aggregator will generically sum the selected values per rows and columns.

  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "Sum Agg", 'digits': 0})

  """
  _dflts = {'digits': 0}  # Cannot be changed directly in the class

  name = "Sum Agg"
  keyAgg, key2Agg = 0, None
  push = 'this.keyAgg += parseFloat(record[attributeArray[0]])'
  value = 'return this.keyAgg'
  format = 'return numberWithCommas(x.toFixed(%(digits)s))'


class JsPivotAbsSumAgg(JsPivotSumAgg):
  """
  Aggregator in charge of adding up all the absolute values based on the defined rows.
  This aggregator will generically sum the absolute selected values per rows and columns.

  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "Abs Sum Agg", 'digits': 0})

  """
  name = "Abs Sum Agg"
  push = 'this.keyAgg += Math.abs(parseFloat(record[attributeArray[0]]))'


class JsPivotMaxAgg(JsPivotSumAgg):
  """
  Aggregator in charge of retrieving the maximum value in a recordset.
  This aggregator will retrieve the maximum value for the given rows and columns

  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "Max Agg", 'digits': 0})

  """
  name = "Max Agg"
  keyAgg = '-Infinity'
  push = "this.keyAgg = Math.max(this.keyAgg, parseFloat(record[attributeArray[0]]))"
  value = 'return this.keyAgg'


class JsPivotMinAgg(JsPivotSumAgg):
  """
  Aggregator in charge of retrieving the minimum value in a recordset.
  This aggregator will retrieve the minimum value for the given rows and columns

  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "Min Agg", 'digits': 0})

  """
  name = "Min Agg"
  keyAgg = 'Infinity'
  push = "this.keyAgg = Math.min(this.keyAgg, parseFloat(record[attributeArray[0]]))"
  value = 'return this.keyAgg'


class JsPivotAvgAgg(JsPivotSumAgg):
  """
  Aggregator in charge of computing the average value in a recordset.
  This aggregator will compute the average value for the given rows and columns

  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "Avg Agg", 'digits': 0})

  """
  name = "Avg Agg"
  keyAgg = 0
  push = "this.keyAgg += parseFloat(record[attributeArray[0]]); this.tmpVal += 1"
  value = "return this.keyAgg / this.tmpVal"


# -------------------------------------------------------------------------------------------------------------------
#                           COMPLEX TRANSFORMATION USING TWO KEYS
#
class JsPivotDiff(JsPivotAggFnc):
  """
  Aggregator in charge of producing the difference.

  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "diff Agg", 'digits': 0})

  """
  name = "diff Agg"
  keyAgg, key2Agg = 0, 0
  push = 'this.key1Agg += parseFloat(record[attributeArray[0]]); this.key2Agg += parseFloat(record[attributeArray[1]])'
  value = 'return this.key1Agg - this.key2Agg'
  format = 'return numberWithCommas(x.toFixed(2))'


class JsPivotDiffPct(JsPivotDiff):
  """
  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "diff Pct Agg", 'digits': 0})

  """
  name = "diff Pct Agg"
  value = 'return (this.key1Agg - this.key2Agg) / this.key1Agg'


class JsPivotDiffAbs(JsPivotDiff):
  """
  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "diff Abs Agg", 'digits': 0})

  """
  name = "diff Abs Agg"
  value = 'return Math.abs(this.key1Agg - this.key2Agg)'


class JsPivotDiffPctAbs(JsPivotDiff):
  """
  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "diff Abs Pct Agg", 'digits': 0})
  """
  name = "diff Abs Pct Agg"
  value = 'return Math.abs((this.key1Agg - this.key2Agg) / this.key1Agg)'


class JsPivotSumOverSumAgg(JsPivotDiff):
  """
  Example
  report.pivot(df, rows=['Date'], cols=['direction'], valCol=['AAPL.Low'], aggOptions={'name': "sum Over Sum Agg", 'digits': 0})
  """
  name = "sum Over Sum Agg"
  push = "this.key1Agg += parseFloat(record[attributeArray[0]]); this.key2Agg += parseFloat(record[attributeArray[1]])"
  value = 'return this.key1Agg / this.key2Agg'




from epyk.core.html import graph


class ChartJs(object):

  def __init__(self, context):
    self.parent = context
    self.chartFamily = "ChartJs"

  def line(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    Documentation
    http://www.chartjs.org/

    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": x, "y": y})
      data.append(series)

    line_chart = graph.GraphChartJs.ChartLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for d in data:
      line_chart.add_dataset(d)

    self.parent.context.register(line_chart)
    return line_chart

  def timeseries(self, record, y_columns=None, x_axis=None, title=None, profile=None, options=None,
                 width=(100, "%"), height=(330, "px"), htmlCode=None):
    line = self.line(record, y_columns, x_axis, profile, width, height, options, htmlCode)
    return line

  def pie(self, record=None, y_column=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param record:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      agg_data[rec[x_axis]] = agg_data.get(rec[x_axis],  0) + float(rec[y_column])
    labels, data = [], []
    for x, y in agg_data.items():
      labels.append(x)
      data.append(y)

    line_chart = graph.GraphChartJs.ChartPie(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.add_dataset(data)
    line_chart.labels(labels)
    self.parent.context.register(line_chart)
    return line_chart

  def donut(self, record=None, y_column=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param record:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      agg_data[rec[x_axis]] = agg_data.get(rec[x_axis],  0) + float(rec[y_column])
    labels, data = [], []
    for x, y in agg_data.items():
      labels.append(x)
      data.append(y)

    dflt_options = {'cutoutPercentage': 50}
    if options is not None:
      dflt_options.update()
    pie_chart = graph.GraphChartJs.ChartPie(self.parent.context.rptObj, width, height, htmlCode, dflt_options, profile)
    pie_chart.add_dataset(data)
    pie_chart.labels(labels)

    self.parent.context.register(pie_chart)
    return pie_chart

  def area(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": x, "y": y})
      data.append(series)

    line_chart = graph.GraphChartJs.ChartLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for d in data:
      line_chart.add_dataset(d, opacity=0.2)
      line_chart.dataset().fill = True
    self.parent.context.register(line_chart)
    return line_chart

  def step(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": x, "y": y})
      data.append(series)

    line_chart = graph.GraphChartJs.ChartLine(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for d in data:
      line_chart.add_dataset(d)
      line_chart.dataset().steppedLine = 'before'
    self.parent.context.register(line_chart)
    return line_chart

  def bar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param record:
    :param y_column:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": x, "y": y})
      data.append(series)

    bar_chart = graph.GraphChartJs.ChartBar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    bar_chart.labels(sorted(list(labels)))
    for d in data:
      bar_chart.add_dataset(d)
    self.parent.context.register(bar_chart)
    return bar_chart

  def hbar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param record:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": y, "y": x})
      data.append(series)

    bar_chart = graph.GraphChartJs.ChartHBar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    bar_chart.labels(sorted(list(labels)))
    for d in data:
      bar_chart.add_dataset(d)
    self.parent.context.register(bar_chart)
    return bar_chart

  def multi(self, data=None, y_columns=None, x_axis=None, title=None, filters=None, profile=None, options=None,
           width=(100, "%"), height=(330, "px"), htmlCode=None):
    """

    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    bar_chart = graph.GraphChartJs.ChartBar(self.parent.context.rptObj, width, height, title, options or {}, htmlCode,
                                              filters, profile)
    bar_chart.chart._data = self.parent.context.rptObj.js.data.records(data).to.chartJs.line(y_columns, x_axis, profile or False)
    self.parent.context.register(bar_chart)
    return bar_chart

  def scatter(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param record:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": x, "y": y})
      data.append(series)

    line_chart = graph.GraphChartJs.ChartScatter(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for d in data:
      line_chart.add_dataset(d)
    self.parent.context.register(line_chart)
    return line_chart

  def bubble(self, record, y_columns=None, x_axis=None, r_values=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data, agg_r = {}, {}
    for rec in record:
      for i, y in enumerate(y_columns):
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
        if r_values is not None and i < len(r_values):
          agg_r.setdefault(y, {})[rec[x_axis]] = agg_r.get(y, {}).get(rec[x_axis],  0) + float(rec[r_values[i]])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append({"x": x, "y": y, 'r': agg_r.get(c, {}).get(x, 1)})
      data.append(series)

    line_chart = graph.GraphChartJs.ChartBubble(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    line_chart.labels(sorted(list(labels)))
    for d in data:
      line_chart.add_dataset(d)
    self.parent.context.register(line_chart)
    return line_chart

  def polar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """

    :param title:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    polar_chart = graph.GraphChartJs.ChartPolar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    polar_chart.labels(sorted(list(labels)))
    for d in data:
      polar_chart.add_dataset(d)
    self.parent.context.register(polar_chart)
    return polar_chart

  def radar(self, record, y_columns=None, x_axis=None, profile=None, width=(100, "%"), height=(330, "px"), options=None, htmlCode=None):
    """
    A radar chart is a way of showing multiple data points and the variation between them.

    https://www.chartjs.org/docs/latest/charts/radar.html

    :param record:
    :param y_columns:
    :param x_axis:
    :param profile:
    :param width:
    :param height:
    :param htmlCode:
    """
    agg_data = {}
    for rec in record:
      for y in y_columns:
        if y in rec:
          agg_data.setdefault(y, {})[rec[x_axis]] = agg_data.get(y, {}).get(rec[x_axis],  0) + float(rec[y])
    labels, data = set(), []
    for c in y_columns:
      series = []
      for x, y in agg_data[c].items():
        labels.add(x)
        series.append(y)
      data.append(series)

    radar_chart = graph.GraphChartJs.ChartRadar(self.parent.context.rptObj, width, height, htmlCode, options, profile)
    radar_chart.labels(sorted(list(labels)))
    for i, d in enumerate(data):
      radar_chart.add_dataset(d)
      radar_chart.dataset().label = y_columns[i]
    self.parent.context.register(radar_chart)
    return radar_chart




from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.core.js.fncs import JsFncsRecords

jsObj = Js.JsBase()
jsObj.registerFunction("row-buckets")
jsObj.addKeyEvent(jsObj.window.alert("test alert"), 13)

f = JsUtils.JsFile("RowBucket", path=r"../outs")
f.writeReport(jsObj)
f.close()






from epyk.core import html
from epyk.core.css import Defaults_css


class Navigation(object):
  def __init__(self, context):
    self.context = context

  def up(self, icon="fas fa-arrow-up", top=20, right=20, bottom=None, tooltip=None, width=(100, '%'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.navigation.up()

    Attributes:
    ----------
    :param icon:
    :param top:
    :param right:
    :param bottom:
    :param tooltip:
    :param width:
    :param options:
    :param profile:
    """
    du = self.context.rptObj.ui.icon(icon).css({"border": '1px solid black', "position": 'fixed', "width": 'none',
                                                "border-radius": '20px', "padding": '8px', "right": '%spx' % right})
    if top is not None:
      du.style.css.top = top
    else:
      du.style.css.bottom = bottom
    du.style.add_classes.div.background_hover()
    self.context.rptObj.js.addOnReady(
      self.context.rptObj.js.window.events.addScrollListener([
        self.context.rptObj.js.if_(self.context.rptObj.js.window.scrollY > 50, [du.dom.show()]).else_(du.dom.hide())
      ]))
    if tooltip is not None:
      du.tooltip(tooltip)
    du.click([
      self.context.rptObj.js.window.scrollUp(),
      self.context.rptObj.js.objects.this.hide()])
    return du

  def down(self, icon="fas fa-arrow-down", top=20, right=20, bottom=None, tooltip=None, width=(100, '%'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.navigation.down()

    Attributes:
    ----------
    :param icon:
    :param top:
    :param right:
    :param bottom:
    :param tooltip:
    :param width:
    :param options:
    :param profile:
    """
    dd = self.context.rptObj.ui.icon(icon).css({"border": '1px solid black', "position": 'fixed', "width": 'none',
        "border-radius": '20px', "padding": '8px', "right": '%spx' % right})
    if bottom is not None:
      dd.style.css.bottom = bottom
    else:
      dd.style.css.top = top
    dd.style.add_classes.div.background_hover()
    self.context.rptObj.js.addOnReady(
      self.context.rptObj.js.window.events.addScrollListener([
        self.context.rptObj.js.if_(self.context.rptObj.js.window.scrollY < (self.context.rptObj.js.window.scrollMaxY - 50), [dd.dom.show()]).else_(dd.dom.hide())
      ]))
    if tooltip is not None:
      dd.tooltip(tooltip)
    dd.click([
      self.context.rptObj.js.window.scrollTo(),
      self.context.rptObj.js.objects.this.hide()])
    return dd

  def to(self, y, x=None, icon="fas fa-map-pin", top=20, right=20, bottom=None, tooltip=None, width=(100, '%'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.navigation.to(100, tooltip="test")

    Attributes:
    ----------
    :param y:
    :param x:
    :param icon:
    :param top:
    :param right:
    :param bottom:
    :param tooltip:
    :param width:
    :param options:
    :param profile:
    """
    dd = self.context.rptObj.ui.icon(icon).css({"border": '1px solid black', "position": 'fixed', "width": 'none',
                                                "border-radius": '20px', "padding": '8px', "right": '%spx' % right})
    if bottom is not None:
      dd.style.css.bottom = bottom
    else:
      dd.style.css.top = top
    dd.style.add_classes.div.background_hover()
    if tooltip is not None:
      dd.tooltip(tooltip)
    self.context.rptObj.js.addOnReady(
      self.context.rptObj.js.window.events.addScrollListener([
        self.context.rptObj.js.if_(self.context.rptObj.js.window.scrollY > y, [dd.dom.show()]).else_(dd.dom.hide())
      ]))
    dd.click([
      self.context.rptObj.js.window.scrollTo(x=x, y=y),
      self.context.rptObj.js.objects.this.hide()])
    return dd

  def pin(self, text, url="#", icon="fas fa-map-pin", top=20, right=20, bottom=None, tooltip=None, width=(100, '%'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.navigation.pin("anchor", tooltip="test", bottom=20)

    Attributes:
    ----------
    :param text:
    :param url:
    :param icon:
    :param top:
    :param right:
    :param bottom:
    :param tooltip:
    :param width:
    :param options:
    :param profile:
    """
    dd = self.context.rptObj.ui.icon(icon)
    h_url = self.context.rptObj.ui.link(text, url=url).css({"margin-left": "10px"})
    div = self.context.rptObj.ui.div([dd, h_url]).css({"border": '1px solid black', "position": 'fixed', "width": 'none',
                          "border-radius": '30px', "padding": '10px 15px', "right": '%spx' % right,
                          "background-color": self.context.rptObj.theme.greys[0]})
    if bottom is not None:
      div.style.css.bottom = bottom
    else:
      div.style.css.top = top
    div.attr['class'].add("CssDivOnHoverWidth")
    h_url.css({"display": 'none', "white-space": 'nowrap'})
    div.on("mouseover", [h_url.dom.css({"display": 'inline-block'})])
    div.on("mouseout", [h_url.dom.css({"display": 'none'})])
    if tooltip is not None:
      div.tooltip(tooltip)
    return div

  def scroll(self, position=0, height=(3, 'px'), options=None, profile=False):
    """
    Description:
    ------------
    Add a horizontal progressbar to display the status of the page scrollbar.

    Usage::

      rptObj.ui.navigation.scroll()

    Attributes:
    ----------
    :param position:
    :param height:
    :param options:
    :param profile:
    """
    p = self.context.rptObj.ui.sliders.progressbar(position, height=height, options=options, profile=profile)
    self.context.rptObj.js.addOnReady(
      self.context.rptObj.js.window.events.addScrollListener([
        p.build(self.context.rptObj.js.window.scrollPercentage)]))
    return p

  def indices(self, count, selected=1, width=(100, '%'), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.navigation.indices(10)

    Attributes:
    ----------
    :param count:
    :param selected:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    dflt_options = {"div_css": {"display": "inline-block", "margin": "0 2px"}, "selected": selected}
    dflt_options.update(options or {})
    html_indices = html.HtmlContainer.Indices(self.context.rptObj, count, width, height, None, dflt_options, profile)
    self.context.register(html_indices)
    return html_indices

  def points(self, count, selected=0, width=(100, '%'), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      p = rptObj.ui.navigation.points(10)
      for i, _ in enumerate(p):
        p.click_item(i, [])

    Attributes:
    ----------
    :param count:
    :param selected:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    dflt_options = {"div_css": {"display": "inline-block", "margin": "0 2px"}, "selected": selected}
    dflt_options.update(options or {})
    html_points = html.HtmlContainer.Points(self.context.rptObj, count, width, height, None, dflt_options, profile)
    self.context.register(html_points)
    return html_points

  def dots(self, count, selected=1, position="right", width=(100, '%'), height=(None, 'px'), options=None,
           profile=False):
    """
    Description:
    ------------

    Usage::

      d = rptObj.ui.navigation.dots(10)

    Attributes:
    ----------
    :param count:
    :param selected:
    :param position:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    dflt_options = {"div_css": {"margin": "2px", "float": position}, "selected": selected}
    dflt_options.update(options or {})
    html_points = html.HtmlContainer.Points(self.context.rptObj, count, width, height, None, dflt_options, profile)
    self.context.register(html_points)
    return html_points

  def path(self, records, divider=None, width=(100, '%'), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      record = [{"text": "Lin 1", 'url': 'report_list.html'}, {"text": "Link 2"}]
      rptObj.ui.navigation.path(record)

    Attributes:
    ----------
    :param records:
    :param divider:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    if divider is None:
      divider = self.context.rptObj.symbols.shapes.BLACK_RIGHT_POINTING_TRIANGLE
    div = self.context.rptObj.ui.div(width=width, height=height, options=options, profile=profile)
    for rec in records[:-1]:
      div += self.context.rptObj.ui.link(rec['text'], url=rec.get('url', '#')).css({"display": 'inline-block'})
      div += self.context.rptObj.ui.text(divider).css({"display": 'inline-block', 'margin': '0 5px', 'font-size': Defaults_css.font(-2)})
    div +=self.context.rptObj.ui.link(records[-1]['text'], url=records[-1].get('url', '#')).css({"display": 'inline-block'})
    return div

  def bar(self, icon=None, title=None, width=(100, '%'), height=(40, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Usage::

      nav = rptObj.ui.navigation.bar(title="test")
      nav.add_text("Test text")
      nav + rptObj.ui.button("Click")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMenu.HtmlNavBar`


    Attributes:
    ----------
    :param icon:
    :param title:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    components = []
    if icon is None:
      components.append(self.context.rptObj.ui.icons.epyk())
    if title is not None:
      title = self.context.rptObj.ui.div(title, height=(100, "%"))
      title.style.css.text_transform = "uppercase"
      title.style.css.margin_left = 5
      title.style.css.bold()
      components.append(title)
    components.append(self.context.rptObj.ui.navigation.scroll())
    html_nav = html.HtmlMenu.HtmlNavBar(self.context.rptObj, components, width=width, height=height, options=options, profile=profile)
    html_nav.style.css.line_height = height[0]
    #html_nav.val[-1].style.css.display = 'block' # scroll object must be block to keep the height, this screws up the display
    self.context.rptObj.body.style.css.padding_top = height[0]
    self.context.register(html_nav)
    return html_nav

  def banner(self, image, text, link, width=(100, '%'), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlImage.Image`
      - :class:`epyk.core.html.HtmlContainer.Col`
      - :class:`epyk.core.html.HtmlContainer.Row`
      - :class:`epyk.core.html.HtmlText.Text`
      - :class:`epyk.core.html.HtmlLinks.ExternalLink`

    :param image:
    :param text:
    :param link:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    div = self.context.rptObj.ui.div(width=width, height=height, options=options, profile=profile)
    h_image = self.context.rptObj.ui.img(image)
    h_text = self.context.rptObj.ui.text(text)
    h_link = self.context.rptObj.ui.links.button("click", link)
    h_row = self.context.rptObj.ui.row(
      [h_image, self.context.rptObj.ui.col([h_text, h_link])])
    div + h_row
    div.style.css.background_color = self.context.rptObj.theme.colors[3]
    div.style.css.color = "white"
    div.style.css.font_size = Defaults_css.font(5)
    div.style.css.text_align = 'center'
    div.style.css.padding = "5px 15px"
    return div

  def footer(self, components=None, width=(100, '%'), height=('120', 'px'), profile=False):
    """
    Description:
    ------------

    Will create a footer object in the body of the report

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMenu.HtmlFooter`

    :param components: list of html components
    :param width: the width of the object
    :param height: the height of the object
    :param profile: get profiling info
    :return:
    """
    footer = html.HtmlMenu.HtmlFooter(self.context.rptObj, components, width=width, height=height, profile=profile)
    self.context.register(footer)
    return footer


class Banners(object):

  def __init__(self, context):
    self.context = context

  def top(self, data, background=None, width=(100, '%'), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`

    :param data:
    :param background:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    div = self.context.rptObj.ui.div(data, width=width, height=height, options=options, profile=profile)
    div.style.css.background_color = background or self.context.rptObj.theme.colors[3]
    div.style.css.color = "white"
    div.style.css.position = "fixed"
    div.style.css.top = 0
    div.style.css.padding = "5px 15px"
    return div

  def bottom(self, data, background=None, width=(100, '%'), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`

    :param data:
    :param background:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    div = self.context.rptObj.ui.div(data, width=width, height=height, options=options, profile=profile)
    div.style.css.background_color = background or self.context.rptObj.theme.colors[3]
    div.style.css.color = "white"
    div.style.css.position = "fixed"
    div.style.css.padding = "5px 15px"
    div.style.css.bottom = 0
    return div

  def corner(self, data, background=None, position="bottom", width=(120, 'px'), height=(None, 'px'), options=None, profile=False):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`

    :param data:
    :param background:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    div = self.context.rptObj.ui.div(data, width=width, height=height, options=options, profile=profile)
    div.style.css.background_color = background or self.context.rptObj.theme.colors[3]
    div.style.css.color = "white"
    div.style.css.position = "fixed"
    div.style.css.padding = "5px 15px"
    div.style.css.text_align = "center"
    div.style.css.right = 0
    if position == 'bottom':
      div.style.css.bottom = 0
      div.style.css.transform = "rotate(-40deg)"
      div.style.css.margin = "0 -30px 15px 0"
    else:
      div.style.css.top = 0
      div.style.css.transform = "rotate(40deg)"
      div.style.css.margin = "15px -30px 0 0"
    return div




from epyk.core.js import Imports

from epyk.core.css.catalogs import CatalogButton
from epyk.core.css.catalogs import CatalogInput
from epyk.core.css.catalogs import CatalogChart
from epyk.core.css.catalogs import CatalogDiv
from epyk.core.css.catalogs import CatalogTree
from epyk.core.css.catalogs import CatalogIcons
from epyk.core.css.catalogs import CatalogLayout
from epyk.core.css.catalogs import CatalogLink
from epyk.core.css.catalogs import CatalogList
from epyk.core.css.catalogs import CatalogPopup
from epyk.core.css.catalogs import CatalogRadio
from epyk.core.css.catalogs import CatalogSelect
from epyk.core.css.catalogs import CatalogTable
from epyk.core.css.catalogs import CatalogText
from epyk.core.css.catalogs import CatalogImg


class Catalog(object):

  def __init__(self, report, classList):
    self.__rptObj, self.__class_list = report, classList
    self.__ctx = {}

  def _class_type(self, type):
    """
    Description:
    ------------

    :param type:

    :return:
    """
    self.__class_list_type = self.__class_list[type]
    return self

  @property
  def button(self):
    """
    Description:
    ------------
    CSS Classes specific to the buttons components

    :rtype: CatalogButton.CatalogButton
    """
    if "button" not in self.__ctx:
      self.__ctx['button'] = CatalogButton.CatalogButton(self.__rptObj, self.__class_list_type)
    return self.__ctx['button']

  @property
  def layout(self):
    """
    Description:
    ------------

    :rtype: CatalogLayout.CatalogLayout
    """
    if "layout" not in self.__ctx:
      self.__ctx['layout'] = CatalogLayout.CatalogLayout(self.__rptObj, self.__class_list_type)
    return self.__ctx['layout']

  @property
  def dropdown(self):
    """
    Description:
    ------------
    CSS Classes specific to the dropdown components

    :rtype: CatalogTree.CssStyleDropdown
    """
    if "dropdown" not in self.__ctx:
      self.__ctx['dropdown'] = CatalogTree.CatalogDropDown(self.__rptObj, self.__class_list_type)
    return self.__ctx['dropdown']

  @property
  def table(self):
    """
    Description:
    ------------
    CSS Classes specific to the Table components

    :rtype: CatalogTable.CatalogTable
    """
    if "table" not in self.__ctx:
      self.__ctx['table'] = CatalogTable.CatalogTable(self.__rptObj, self.__class_list_type)
    return self.__ctx['table']

  @property
  def chart(self):
    """
    Description:
    ------------
    CSS Classes specific to the buttons components

    :rtype: CatalogButton.CatalogButton
    """
    if "chart" not in self.__ctx:
      self.__ctx['chart'] = CatalogChart.CatalogChart(self.__rptObj, self.__class_list_type)
    return self.__ctx['chart']

  @property
  def link(self):
    """
    Description:
    ------------

    :rtype: CatalogLink.CatalogLink
    """
    if "link" not in self.__ctx:
      self.__ctx['link'] = CatalogLink.CatalogLink(self.__rptObj, self.__class_list_type)
    return self.__ctx['link']

  @property
  def date(self):
    """
    Description:
    ------------

    :rtype: CatalogInput.CatalogDate
    """
    if "date" not in self.__ctx:
      self.__ctx['date'] = CatalogInput.CatalogDate(self.__rptObj, self.__class_list_type)
    return self.__ctx['date']

  @property
  def text(self):
    """
    Description:
    ------------

    :rtype: CatalogText.CatalogText
    """
    if "text" not in self.__ctx:
      self.__ctx['text'] = CatalogText.CatalogText(self.__rptObj, self.__class_list_type)
    return self.__ctx['text']

  @property
  def input(self):
    """
    Description:
    ------------
    """
    if "inputs" not in self.__ctx:
      self.__ctx['inputs'] = CatalogInput.CatalogInput(self.__rptObj, self.__class_list_type)
    return self.__ctx['inputs']

  @property
  def image(self):
    """
    """
    if "image" not in self.__ctx:
      self.__ctx['image'] = CatalogImg.CatalogImg(self.__rptObj, self.__class_list_type)
    return self.__ctx['image']

  @property
  def div(self):
    """
    Description:
    ------------

    :rtype: CatalogDiv.CatalogDiv
    """
    if "div" not in self.__ctx:
      self.__ctx['div'] = CatalogDiv.CatalogDiv(self.__rptObj, self.__class_list_type)
    return self.__ctx['div']

  @property
  def radio(self):
    """
    Description:
    ------------

    :rtype: CatalogRadio.CatalogRadio
    """
    if "radio" not in self.__ctx:
      self.__ctx['radio'] = CatalogRadio.CatalogRadio(self.__rptObj, self.__class_list_type)
    return self.__ctx['radio']

  def custom(self, cssClass):
    """
    Description:
    ------------
    Register a bespoke external class to the report object

    Attributes:
    ----------
    :param cssClass:

    :return:
    """
    cssObj = cssClass(self.__rptObj)
    cssObj.customize()
    self.__class_list_type.add(cssObj)
    return self

  def customFile(self, filename, path=None):
    """
    Description:
    ------------
    Add an external CSS file to the final HTML report

    Attributes:
    ----------
    :param filename: String. The file name
    :param path: String. Optional. The full path of the external CSS file. If None the user part in Imports.STATIC_PATH
                 will be used

    :return:
    """
    if path is None:
      self.__rptObj.cssLocalImports.add("%s/css/%s" % (Imports.STATIC_PATH.replace("\\", "/"), filename))
    else:
      self.__rptObj.cssLocalImports.add("%s/%s" % (path, filename))
    return self

  def customText(self, text):
    """
    Description:
    ------------
    Add a bespoke CSS fragment

    Attributes:
    ----------
    :param text: String. The CSS fragment to be added to the HTML report. THis can be a class or a group of class

    :return:
    """
    self.__rptObj._cssText.append(text)
    return self

  def anonymous_cls(self, attrs):
    """
    Description:
    ------------
    Create a bespoke class based on the various attributes.
    This will internal build the class and return it

    Usage::

      v_cls = rptObj.css.anonymous_cls({
      '_attrs': {'color': 'green', 'cursor': 'pointer'},
      '_hover': {'color': 'red'}})

    Attributes:
    ----------
    :param attrs: Dictionary. The expected class attributes

    :return: The Python class
    """
    import hashlib

    from epyk.core.css.styles.classes import CssStyle

    has_style = str(hashlib.sha1(str(attrs).encode()).hexdigest())
    attrs['classname'] = "style_%s" % has_style
    meta_cls = type('Style%s' % has_style, (CssStyle.Style,), attrs)
    return meta_cls

  def external(self, classname):
    """
    Description:
    ------------
    Add external CSS classes to a component

    :param classname: String or array

    :return:
    """
    if isinstance(classname, list):
      classname = " ".join(classname)
    self.__class_list_type.add(classname)



"""
Interface to the SipHash module available in the below packages
Related Pages:

			https://pypi.org/project/siphash/
"""


class SipHash(object):
    """
    Generate a unique hash ID from the given string.
    This is supposed to be unique with a minimum expectation of collisions.
    This module is only in charge of producing the hash ID and the potential collisions should be monitored in the environment by the users
    """
    def __init__(self, c=2, d=4):
      assert c >= 0
      assert d >= 0
      self.__c = c
      self.__d = d

    def __process_message(self, message):
      self.__message = []

      length = len(message)
      for start in range(0, length - 7, 8):
          state = ord(message[start]) \
              | (ord(message[start + 1]) << 8)  \
              | (ord(message[start + 2]) << 16) \
              | (ord(message[start + 3]) << 24) \
              | (ord(message[start + 4]) << 32) \
              | (ord(message[start + 5]) << 40) \
              | (ord(message[start + 6]) << 48) \
              | (ord(message[start + 7]) << 56)
          self.__message.append(state)

      start = (length // 8) * 8
      state = (length % 256) << 56
      for i in range(length - start):
          state |= (ord(message[start + i]) << (i * 8))
      self.__message.append(state)

    def __SipRound(self):
      self.__v0 += self.__v1  # no need to mod 2^64 now
      self.__v2 += self.__v3
      self.__v1 = (self.__v1 << 13) | (self.__v1 >> 51)
      self.__v3 = (self.__v3 << 16) | (self.__v3 >> 48)
      self.__v1 ^= self.__v0
      self.__v3 ^= self.__v2
      self.__v0 = (self.__v0 << 32) | ((self.__v0 >> 32) & 0xffffffff)
      self.__v2 += self.__v1
      self.__v0 += self.__v3
      self.__v0 &= 0xffffffffffffffff
      self.__v1 = (self.__v1 << 17) | ((self.__v1 >> 47) & 0x1ffff)
      self.__v3 = ((self.__v3 & 0x7ffffffffff) << 21) \
          | ((self.__v3 >> 43) & 0x1fffff)
      self.__v1 ^= self.__v2
      self.__v1 &= 0xffffffffffffffff
      self.__v3 ^= self.__v0
      self.__v2 = ((self.__v2 & 0xffffffff) << 32) \
          | ((self.__v2 >> 32) & 0xffffffff)

    def auth(self, key, message):
      assert 0 <= key and key < 1 << 128
      k0 = key & 0xffffffffffffffff
      k1 = key >> 64

      # initialization
      self.__v0 = k0 ^ 0x736f6d6570736575
      self.__v1 = k1 ^ 0x646f72616e646f6d
      self.__v2 = k0 ^ 0x6c7967656e657261
      self.__v3 = k1 ^ 0x7465646279746573

      self.__process_message(message)

      # compression
      for m in self.__message:
          self.__v3 ^= m
          for i in range(self.__c):
              self.__SipRound()
          self.__v0 ^= m

      # finalization
      self.__v2 ^= 0xff
      for i in range(self.__d):
          self.__SipRound()
      return self.__v0 ^ self.__v1 ^ self.__v2 ^ self.__v3

    def hashId(self, text):
      """
      Produce a unique ID for a given string. This can be used to replace the internal numbers.

      Example
      >>> SipHash().hashId("Test")
      3169529217224722230

      Related Pages:

			https://github.com/bozhu/siphash-python

      :param text: The String to be hashed
      """
      return self.auth(0x0f0e0d0c0b0a09080706050403020100, text)





from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesDrop
from epyk.core.css.styles.classes import CssStylesDivEvents
from epyk.core.css.styles.classes import CssStylesDiv
from epyk.core.css.styles.classes import CssStylesCommon
from epyk.core.css.styles.classes import CssStylesDivDrawers


class CatalogDiv(Catalog.CatalogGroup):
  def basic(self):
    """  """
    return self._set_class(CssStylesDrop.CssDropFile)

  def bubble(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivBubble)

  def bubble_comment(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivCommBubble)

  def left(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivLeft)

  def right(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivRight)

  def pointer(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivCursor)

  def banner(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivBanner)

  def sub_banner(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivSubBanner)

  def sub_label(self):
    """ """
    return self._set_class(CssStylesDiv.CssDivLabelPoint)

  def no_border(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivNoBorder)

  def border(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivWithBorder)

  def border_shadow(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivShadow)

  def border_dot(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivBoxWithDotBorder)

  def border_bottom(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivBottomBorder)

  def console(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivConsole)

  def margin_with_border(self):
    """  """
    return self._set_class(CssStylesDiv.CsssDivBoxMarginBorder)

  def no_margin(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivNoMargin)

  def margin_vertical(self):
    """  """
    return self._set_class(CssStylesDiv.CsssDivBoxMarginVertical)

  def content_center(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivBoxCenter)

  def color_hover(self):
    """ Change the color when the mouse is on the component """
    return self._set_class(CssStylesDivEvents.CssDivOnHover)

  def background_hover(self):
    """ Change the background color when the mouse is on the component """
    return self._set_class(CssStylesDivEvents.CssDivOnHoverBackgroundLight)

  def width_hover(self):
    """ Change the background color when the mouse is on the component """
    return self._set_class(CssStylesDivEvents.CssDivOnHoverWidth)

  def span_close(self):
    """Change the font size and location of the close button - generally used for the modal components"""
    return self._set_class(CssStylesCommon.CssCloseSpan)

  def modal(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivModal)

  def modal_content(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivModalContent)

  def stepper(self):
    """  """
    return self._set_class(CssStylesDiv.CssDivStepper)


class CatalogDrawer(Catalog.CatalogGroup):

  def drawer(self):
    """  The main layout for the drawer compenent """
    return self._set_class(CssStylesDivDrawers.CssDrawer)

  def nav(self):
    """   """
    return self._set_class(CssStylesDivDrawers.CssDrawerNav)

  def handle(self):
    """  """
    return self._set_class(CssStylesDivDrawers.CssDrawerHandle)

  def content(self):
    """  """
    return self._set_class(CssStylesDivDrawers.CssDrawerContent)



"""

"""

import json

from epyk.core.js import JsUtils


class JQueryTime(object):
  class __internal(object):
    jqId, htmlId, jsImports, cssImport = '', '', set([]), set([])

  def __init__(self, src=None):
    self.src = src if src is not None else self.__internal()
    self.src.jsImports.add('timepicker')
    self.src.cssImport.add('timepicker')
    #self.selector = self.src.jqId if hasattr(self.src, 'jqId') else None
    self._js = []




from epyk.core.css.styles import GrpCls
from epyk.core.css.styles.attributes import AttrClsButtons

from epyk.core.css import Properties
from epyk.core.css import Classes


class ClassButton(GrpCls.ClassHtml):

  @property
  def css(self):
    """
    Description:
    -----------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsButtons.AttrButton
    """
    if self._css_struct is None:
      self._css_struct = AttrClsButtons.AttrButton(self.htmlObj)
    return self._css_struct

  @property
  def css_class(self):
    """
    Description:
    -----------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogButton.CatalogButton
    """
    if self._css_class is None:
      self._css_class = Classes.CatalogButton.CatalogButton(self.htmlObj._report, self.classList['main'], html_id=self.htmlObj.htmlId).basic()
    return self._css_class


class ClassBadge(GrpCls.ClassHtml):

  @property
  def css(self):
    """
    Description:
    -----------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsButtons.AttrBadge
    """
    if self._css_struct is None:
      self._css_struct = AttrClsButtons.AttrBadge(self.htmlObj)
    return self._css_struct


class ClassButtonCheckBox(GrpCls.ClassHtml):

  @property
  def css(self):
    """
    Description:
    -----------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: AttrClsButtons.AttrButton
    """
    if self._css_struct is None:
      self._css_struct = AttrClsButtons.AttrButton(self.htmlObj)
    return self._css_struct


class ClassButtonMenu(GrpCls.ClassHtml):

  def __init__(self, htmlObj):
    super(ClassButtonMenu, self).__init__(htmlObj)
    self._css_btn_content_hover, self._css_btn_link_hover = None, None
    self.classList['main'].add(self.css_btn_content)
    self.classList['main'].add(self.css_btn_link_hover)

  @property
  def css_btn_content(self):
    """
    Description:
    -----------
    The CSS property of the underlying items panel.
    This component will have a dedicated CSS class for the hover event

    :rtype: Classes.CatalogButton.CatalogButton
    """
    if self._css_btn_content_hover is None:
      self._css_btn_content_hover = Classes.CatalogButton.CatalogButton(self.htmlObj._report, self.classList['main'],
                                                            html_id=self.htmlObj.htmlId).content()
    return self._css_btn_content_hover

  @property
  def css_btn_link_hover(self):
    """
    Description:
    -----------
    The CSS property of the underlying item.
    This component will have a dedicated CSS class for the hover event

    :rtype: Classes.CatalogButton.CatalogButton
    """
    if self._css_btn_link_hover is None:
      self._css_btn_link_hover = Classes.CatalogButton.CatalogButton(self.htmlObj._report, self.classList['main'],
                                                                        html_id=self.htmlObj.htmlId).content_link()
    return self._css_btn_link_hover

# class CssClassButtonCheckBox(CssGrpCls.CssGrpClass):
#   """
#
#   """
#   css_button_basic = CssStylesLabel.CssLabelContainer
#   css_label_container_disabled = CssStylesLabel.CssLabelContainerDisabled
#   css_label_check_mark_hover = CssStylesLabel.CssLabelCheckMarkHover
#   css_div_no_border = CssStylesDiv.CssDivNoBorder
#   css_check_mark = CssStylesText.CssCheckMark
#   __map, __alt_map = ['CssButtonBasic', 'CssLabelCheckMarkHover', 'CssDivNoBorder', 'CssCheckMark',
#                       'CssLabelContainerDisabled'], []
